import aerosandbox.numpy as np
from typing import Union


def CDA_control_linkage(
        Re_l: Union[float, np.ndarray],
        linkage_length: Union[float, np.ndarray],
        is_covered: Union[bool, np.ndarray] = False,
        is_top: Union[bool, np.ndarray] = False,
) -> Union[float, np.ndarray]:
    """
    Computes the drag area (CDA) of a typical control usage as used on a well-manufactured RC airplane.

    The drag area (CDA) is defined as: CDA == D / q, where:

        - D is the drag force (dimensionalized, e.g., in Newtons)

        - q is the freestream dynamic pressure (dimensionalized, e.g., in Pascals)

    See study with original data at `AeroSandbox/studies/LinkageDrag`.

    Data from:

        * Hepperle, Martin. "Drag of Linkages". https://www.mh-aerotools.de/airfoils/linkage.htm

        * Summarizes data from "Werner Würz, published in the papers of the ISF-Seminar in December 1989 in Baden, Switzerland."

    Args:

        Re_l: Reynolds number, with reference length as the length of the linkage.

        linkage_length: The length of the linkage. [m]

        is_covered: A boolean of whether an aerodynamic fairing is placed around the linkage.

        is_top: A boolean of whether the linkage is on the top surface of the wing (True) or the bottom surface (
            False). Differences in local boundary layer and inviscid effects cause local velocity changes.

    Returns: The drag area [m^2] of the control linkage.

    """
    x = dict(
        Re_l=Re_l,
        linkage_length=linkage_length,
        is_covered=is_covered,
        is_top=is_top
    )

    p = {
        'CD0'               : 7.833083680086374e-05,
        'CD1'               : 0.0001216877860785463,
        'c_length'          : 30.572471745477774,
        'covered_drag_ratio': 0.7520722978405192,
        'top_drag_ratio'    : 1.1139040832208857
    }

    Re = x["Re_l"]
    linkage_length = x["linkage_length"]
    is_covered = x["is_covered"]
    is_top = x["is_top"]

    side_drag_multiplier = np.where(
        is_top,
        p["top_drag_ratio"],
        1
    )
    covered_drag_multiplier = np.where(
        is_covered,
        p["covered_drag_ratio"],
        1
    )
    linkage_length_multiplier = 1 + p["c_length"] * linkage_length

    CDA_raw = (
            p["CD1"] / (Re / 1e5) +
            p["CD0"]
    )

    return side_drag_multiplier * covered_drag_multiplier * linkage_length_multiplier * CDA_raw


def CDA_control_surface_gaps(
        local_chord: float,
        control_surface_span: float,
        local_thickness_over_chord: float = 0.12,
        control_surface_hinge_x: float = 0.75,
        n_side_gaps: int = 2,
        side_gap_width: float = None,
        hinge_gap_width: float = None,
) -> float:
    """
    Computes the drag area (CDA) of the gaps associated with a typical wing control surface.
    (E.g., aileron, flap, elevator, rudder).

    The drag area (CDA) is defined as: CDA == D / q, where:

        - D is the drag force (dimensionalized, e.g., in Newtons)

        - q is the freestream dynamic pressure (dimensionalized, e.g., in Pascals)

    This drag area consists of two sources:

        1. Chordwise gaps at the side edges of the control surface ("side gaps")

        2. Spanwise gaps at the hinge line of the control surface ("hinge gap")

    Args:

        local_chord: The local chord of the wing at the midpoint of the control surface. [meters]

        control_surface_span: The span of the control surface. [meters]

        local_thickness_over_chord: The local thickness-to-chord ratio of the wing at the midpoint of the control
            surface. [nondimensional] For example, this is 0.12 for a NACA0012 airfoil.

        control_surface_hinge_x: The x-location of the hinge line of the control surface, as a fraction of the local
            chord. [nondimensional] Defaults to x_hinge / c = 0.75, which is typical for an aileron.

        n_side_gaps: The number of "side gaps" to count on this control surface when computing drag. Defaults to 2 (
            i.e., one inboard gap, one outboard gap), which is the simplest case of a wing with a single partial-span
            aileron. However, there may be cases where it is best to reduce this to 1 or 0. For example:

                * A wing with a single full-span aileron would have 1 side gap (at the wing root, but not at the tip).

                * A wing with a flap and aileron that share a chordwise gap would be best modeled by setting
                    n_side_gaps = 1 ( so that no double-counting occurs).

        side_gap_width: The width of the chordwise gaps at the side edges of the control surface [meters]. If this is
            left as the default (None), then a typical value will be computed based on the local chord and control surface
            span.

        hinge_gap_width: The width of the spanwise gap at the hinge line of the control surface [meters]. If this is
            left as the default (None), then a typical value will be computed based on the local chord.

    Returns: The drag area [m^2] of the gaps associated with the control surface. This should be added to the "clean"
        wing drag to get a more realistic drag estimate.

    """
    if side_gap_width is None:
        side_gap_width = np.maximum(
            np.maximum(
                0.002,
                0.006 * local_chord
            ),
            control_surface_span * 0.01
        )
    if hinge_gap_width is None:
        hinge_gap_width = 0.03 * local_chord

    ### Chordwise gaps (at side edges of control surface)
    """
    Based on Hoerner, "Fluid Dynamic Drag", 1965, p. 5-13. Figure 26, "Drag of longitudinal wing gaps, 
    tested on 2412 airfoil at C_L = 0.1 and Re_c = 2 * 10^6"
    """

    CDA_side_gaps = n_side_gaps * (side_gap_width * local_chord * local_thickness_over_chord) * 0.50

    ### Spanwise gaps (at hinge line of control surface)
    """
    Based on Hoerner, "Fluid Dynamic Drag", 1965, p. 5-13. Figure 27, "Evaluation of drag due to control gaps"
    """

    CDA_hinge_gap = 0.025 * hinge_gap_width * control_surface_span

    ### Total
    return CDA_side_gaps + CDA_hinge_gap


def CDA_protruding_bolt_or_rivet(
        diameter: float,
        kind: str = "flush_rivet"
):
    """
    Computes the drag area (CDA) of a protruding bolt or rivet.

    The drag area (CDA) is defined as: CDA == D / q, where:
        - D is the drag force (dimensionalized, e.g., in Newtons)
        - q is the freestream dynamic pressure (dimensionalized, e.g., in Pascals)

    Args:

        diameter: The diameter of the bolt or rivet. [meters]

        kind: The type of bolt or rivet. Valid options are:

            - "flush_rivet"

            - "round_rivet"

            - "flat_head_bolt"

            - "round_head_bolt"

            - "cylindrical_bolt"

            - "hex_bolt"

    Returns: The drag area [m^2] of the bolt or rivet.
    """
    S_ref = np.pi * diameter ** 2 / 4

    CD_factors = {
        "flush_rivet"     : 0.002,
        "round_rivet"     : 0.04,
        "flat_head_bolt"  : 0.02,
        "round_head_bolt" : 0.32,
        "cylindrical_bolt": 0.42,
        "hex_bolt"        : 0.80,
    }

    try:
        CDA = CD_factors[kind] * S_ref
    except KeyError:
        raise ValueError("Invalid `kind` of bolt or rivet.")

    return CDA


def CDA_perpendicular_sheet_metal_joint(
        joint_width: float,
        sheet_metal_thickness: float,
        kind: str = "butt_joint_with_inside_joiner"
):
    """
    Computes the drag area (CDA) of a sheet metal joint that is perpendicular to the flow.
        (E.g., spanwise on the wing, or circumferential on the fuselage).

    The drag area (CDA) is defined as: CDA == D / q, where:

        - D is the drag force (dimensionalized, e.g., in Newtons)

        - q is the freestream dynamic pressure (dimensionalized, e.g., in Pascals)

    Args:

        joint_width: The width of the joint (perpendicular to the airflow, e.g., spanwise on a wing). [meters]

        sheet_metal_thickness: The thickness of the sheet metal. [meters]

        kind: The type of joint. Valid options are:

            - "butt_joint_with_inside_joiner"

            - "butt_joint_with_inside_weld"

            - "butt_joint_with_outside_joiner"

            - "butt_joint_with_outside_weld"

            - "lap_joint_forward_facing_step"

            - "lap_joint_backward_facing_step"

            - "lap_joint_forward_facing_step_with_bevel"

            - "lap_joint_backward_facing_step_with_bevel"

            - "lap_joint_forward_facing_step_with_rounded_bevel"

            - "lap_joint_backward_facing_step_with_rounded_bevel"

            - "flush_lap_joint_forward_facing_step"

            - "flush_lap_joint_backward_facing_step"


    Returns: The drag area [m^2] of the sheet metal joint.

    """
    S_ref = joint_width * sheet_metal_thickness

    CD_factors = {
        "butt_joint_with_inside_joiner"                    : 0.01,
        "butt_joint_with_inside_weld"                      : 0.01,
        "butt_joint_with_outside_joiner"                   : 0.70,
        "butt_joint_with_outside_weld"                     : 0.51,
        "lap_joint_forward_facing_step"                    : 0.40,
        "lap_joint_backward_facing_step"                   : 0.22,
        "lap_joint_forward_facing_step_with_bevel"         : 0.11,
        "lap_joint_backward_facing_step_with_bevel"        : 0.24,
        "lap joint_forward_facing_step_with_rounded_bevel" : 0.04,
        "lap_joint_backward_facing_step_with_rounded_bevel": 0.16,
        "flush_lap_joint_forward_facing_step"              : 0.13,
        "flush_lap_joint_backward_facing_step"             : 0.07,
    }

    try:
        CDA = CD_factors[kind] * S_ref
    except KeyError:
        raise ValueError("Invalid `kind` of sheet metal joint.")

    return CDA


# def CDA_pitot_static_tube():
#     pass
#
#
# def CDA_landing_gear():
#     pass



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import List
import copy
import pytest

vector = [1, 2, 3]
op_point = asb.OperatingPoint(alpha=10, beta=5)


def chain_conversion(
        axes: List[str] = None
):
    if axes is None:
        axes = ["geometry", "body", "geometry"]

    x, y, z = copy.deepcopy(vector)
    for from_axes, to_axes in zip(axes, axes[1:]):
        x, y, z = op_point.convert_axes(
            x_from=x,
            y_from=y,
            z_from=z,
            from_axes=from_axes,
            to_axes=to_axes
        )
    return np.array(vector) == pytest.approx(np.array([x, y, z]))


def test_basic():
    assert chain_conversion()


def test_geometry():
    assert chain_conversion(["body", "geometry", "body"])


def test_stability():
    assert chain_conversion(["body", "stability", "body"])


def test_wind():
    assert chain_conversion(["body", "wind", "body"])


def test_cycle():
    assert chain_conversion([
        "body",
        "geometry",
        "stability",
        "wind",
        "body",
        "wind",
        "stability",
        "geometry",
        "body",
        "geometry",
        "wind",
        "geometry",
        "stability",
        "body",
    ])


if __name__ == '__main__':
    pytest.main()
    chain_conversion()



import aerosandbox.numpy as np
from aerosandbox.atmosphere.atmosphere import Atmosphere
from typing import Union

"""
Welcome to the AeroSandbox solar energy library!

The function you're probably looking for is `solar_flux()`, which summarizes this entire module and computes the 
realized solar flux on a given surface as a function of many different parameters.
"""


def _prepare_for_inverse_trig(x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
    """
    Ensures that a value is within the open interval (-1, 1), so that if you call an inverse trig function
    on it (e.g., arcsin, arccos), you won't get an infinity or NaN.

    Args:
        x: A floating-point number or an array of such. If an array, this function acts elementwise.

    Returns: A clipped version of the number, constrained to be in the open interval (-1, 1).
    """
    return (
            np.nextafter(1, -1) *
            np.clip(x, -1, 1)
    )


def solar_flux_outside_atmosphere_normal(
        day_of_year: Union[int, float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Computes the normal solar flux at the top of the atmosphere ("Airmass 0").
    This varies due to Earth's orbital eccentricity (elliptical orbit).

    Source: https://www.itacanet.org/the-sun-as-a-source-of-energy/part-2-solar-energy-reaching-the-earths-surface/#2.1.-The-Solar-Constant

    Args:
        day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)

    Returns: The normal solar flux [W/m^2] at the top of the atmosphere.

    """
    return 1367 * (
            1 + 0.034 * np.cosd(360 * (day_of_year) / 365.25)
    )


def declination_angle(
        day_of_year: Union[int, float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Computes the solar declination angle, in degrees, as a function of day of year.

    Accounts for the Earth's obliquity.

    Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/declination-angle

    Args:
        day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)

    Returns: Solar declination angle [deg]

    """
    return -23.4398 * np.cosd(360 / 365.25 * (day_of_year + 10))


def solar_elevation_angle(
        latitude: Union[float, np.ndarray],
        day_of_year: Union[int, float, np.ndarray],
        time: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Elevation angle of the sun [degrees] for a local observer.

    Solar elevation angle is the angle between the Sun's position and the local horizon plane.
    (Solar elevation angle) = 90 deg - (solar zenith angle)

    Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/elevation-angle

    Args:
        latitude: Local geographic latitude [degrees]. Positive for north, negative for south.
        day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
        time: Time after local solar noon [seconds]

    Returns: Solar elevation angle [degrees] (angle between horizon and sun). Returns negative values if the sun is
    below the horizon.

    """
    declination = declination_angle(day_of_year)

    sin_solar_elevation_angle = (
            np.sind(declination) * np.sind(latitude) +
            np.cosd(declination) * np.cosd(latitude) * np.cosd(time / 86400 * 360)
    )

    solar_elevation_angle = np.arcsind(
        _prepare_for_inverse_trig(sin_solar_elevation_angle)
    )  # in degrees
    return solar_elevation_angle


def solar_azimuth_angle(
        latitude: Union[float, np.ndarray],
        day_of_year: Union[int, float, np.ndarray],
        time: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Azimuth angle of the sun [degrees] for a local observer.

    Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/azimuth-angle

    Args:
        latitude: Local geographic latitude [degrees]. Positive for north, negative for south.
        day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
        time: Time after local solar noon [seconds]

    Returns: Solar azimuth angle [degrees] (the compass direction from which the sunlight is coming).
        * 0 corresponds to North, 90 corresponds to East.
        * Output ranges from 0 to 360 degrees.

    """
    declination = declination_angle(day_of_year)
    sdec = np.sind(declination)
    cdec = np.cosd(declination)
    slat = np.sind(latitude)
    clat = np.cosd(latitude)
    ctime = np.cosd(time / 86400 * 360)

    elevation = solar_elevation_angle(latitude, day_of_year, time)
    cele = np.cosd(elevation)

    cos_azimuth = (sdec * clat - cdec * slat * ctime) / cele

    azimuth_raw = np.arccosd(_prepare_for_inverse_trig(cos_azimuth))

    is_solar_morning = np.mod(time, 86400) > 43200

    solar_azimuth_angle = np.where(
        is_solar_morning,
        azimuth_raw,
        360 - azimuth_raw
    )

    return solar_azimuth_angle


def airmass(
        solar_elevation_angle: Union[float, np.ndarray],
        altitude: Union[float, np.ndarray] = 0.,
        method='Young'
) -> Union[float, np.ndarray]:
    """
    Computes the (relative) airmass as a function of the (true) solar elevation angle and observer altitude.
    Includes refractive (e.g. curving) effects due to atmospheric density gradient.

    Airmass is the line integral of air density along an upwards-pointed ray, extended to infinity. As a raw
    calculation of "absolute airmass", this would have units of kg/m^2. It varies primarily as a function of solar
    elevation angle and observer altitude. (Higher altitude means less optical absorption.) However,
    "airmass" usually refers to the "relative airmass", which is the absolute airmass of a given scenario divided by
    the absolute airmass of a reference scenario. This reference scenario is when the sun is directly overhead (solar
    elevation angle of 90 degrees) and the observer is at sea level.

    Therefore:

        * Outer space has a (relative) airmass of 0 (regardless of solar elevation angle).

        * Sea level with the sun directly overhead has a (relative) airmass of 1.

        * Sea level with the sun at the horizon has a (relative) airmass of ~31.7. (Not infinity, since the Earth is
        spherical, so the ray eventually reaches outer space.) Some models will say that this relative airmass at the
        horizon should be ~38; that is only true if one uses the *apparent* solar elevation angle, rather than the
        *true* (geometric) one. The discrepancy comes from the fact that light refracts (curves) as it passes through
        the atmosphere's density gradient, with the difference between true and apparent elevation angles reaching a
        maximum of 0.56 degrees at the horizon.

    Solar elevation angle is the angle between the Sun's position and the horizon.
    (Solar elevation angle) = 90 deg - (solar zenith angle)

    Note that for small negative values of the solar elevation angle (e.g., -0.5 degree), airmass remains finite,
    due to ray refraction (curving) through the atmosphere.

    For significantly negative values of the solar elevation angle (e.g., -10 degrees), the airmass is theoretically
    infinite. This function returns 1e100 in lieu of this here.

    Sources:

        Young model: Young, A. T. 1994. Air mass and refraction. Applied Optics. 33:1108–1110. doi:
        10.1364/AO.33.001108. Reproduced at https://en.wikipedia.org/wiki/Air_mass_(astronomy)

    Args:

        solar_elevation_angle: Solar elevation angle [degrees] (angle between horizon and sun). Note that we use the
        true solar elevation angle, rather than the apparent one. The discrepancy comes from the fact that light
        refracts (curves) as it passes through the atmosphere's density gradient, with the difference between true
        and apparent elevation angles reaching a maximum of 0.56 degrees at the horizon.

        altitude: Altitude of the observer [meters] above sea level.

        method: A string that determines which model to use.

    Returns: The relative airmass [unitless] as a function of the (true) solar elevation angle and observer altitude.

        * Always ranges from 0 to Infinity

    """
    true_zenith_angle = 90 - solar_elevation_angle

    if method == 'Young':
        cos_zt = np.cosd(true_zenith_angle)
        cos2_zt = cos_zt ** 2
        cos3_zt = cos_zt ** 3

        numerator = 1.002432 * cos2_zt + 0.148386 * cos_zt + 0.0096467
        denominator = cos3_zt + 0.149864 * cos2_zt + 0.0102963 * cos_zt + 0.000303978

        sea_level_airmass = np.where(
            denominator > 0,
            numerator / denominator,
            1e100  # Essentially, infinity.
        )
    else:
        raise ValueError("Bad value of `method`!")

    airmass_at_altitude = sea_level_airmass * (
            Atmosphere(altitude=altitude).pressure() /
            101325.
    )

    return airmass_at_altitude


def solar_flux(
        latitude: Union[float, np.ndarray],
        day_of_year: Union[int, float, np.ndarray],
        time: Union[float, np.ndarray],
        altitude: Union[float, np.ndarray] = 0.,
        panel_azimuth_angle: Union[float, np.ndarray] = 0.,
        panel_tilt_angle: Union[float, np.ndarray] = 0.,
        air_quality: str = 'typical',
        albedo: Union[float, np.ndarray] = 0.2,
        **deprecated_kwargs
) -> Union[float, np.ndarray]:
    """
    Computes the solar power flux (power per unit area) on a flat (possibly tilted) panel. Accounts for atmospheric
    absorption, scattering, and re-scattering (e.g. diffuse illumination), all as a function of panel altitude.

    Fully vectorizable.

    Source for atmospheric absorption:

        * Planning and installing photovoltaic systems: a guide for installers, architects and engineers,
        2nd Ed. (2008), Table 1.1, Earthscan with the International Institute for Environment and Development,
        Deutsche Gesellschaft für Sonnenenergie. ISBN 1-84407-442-0., accessed via
        https://en.wikipedia.org/wiki/Air_mass_(solar_energy)

    Args:

        latitude: Local geographic latitude [degrees]. Positive for north, negative for south.

        day_of_year: The day of the year, represented in the Julian day format (i.e., 1 == Jan. 1, 365 == Dec. 31). This
            accounts for seasonal variations in the sun's position in the sky.

        time: The time of day, measured as the time elapsed after local solar noon [seconds]. Should range from 0 to
            86,400 (the number of seconds in a day). Local solar noon is the time of day when the sun is at its highest
            point in the sky, directly above the observer's local meridian. This is the time when the sun's rays are most
            directly overhead and solar flux is at its peak for a given location. Solar noon does not necessarily occur
            at exactly 12:00 PM local standard time, as it depends on your longitude, the equation of time, and the time
            of year. (Also, don't forget to account for Daylight Savings Time, if that's a relevant consideration for
            your location and season.) Typically, local solar noon is +- 15 minutes from 12:00 PM local standard time.

        altitude: Altitude of the panel above sea level [meters]. This affects atmospheric absorption and scattering
            characteristics.

        panel_azimuth_angle: The azimuth angle of the panel normal [degrees] (the compass direction in which the
            panel normal is tilted). Irrelevant if the panel tilt angle is 0 (e.g., the panel is horizontal).

            * 0 corresponds to North, 90 corresponds to East.

            * Input ranges from 0 to 360 degrees.

        panel_tilt_angle: The angle between the panel normal and vertical (zenith) [degrees].

            * Note: this angle convention is different than the solar elevation angle convention!

            * A horizontal panel has a tilt angle of 0, and a vertical panel has a tilt angle of 90 degrees.

            If the angle between the panel normal and the sun direction is ever more than 90 degrees (e.g. the panel
            is pointed the wrong way), we assume that the panel receives no direct irradiation. (However,
            it may still receive minor amounts of power due to diffuse irradiation from re-scattering.)

        air_quality: Indicates the amount of pollution in the air. A string, one of:

            * 'clean': Pristine atmosphere conditions.
            
            * 'typical': Corresponds to "rural aerosol loading" following ASTM G-173.

            * 'polluted': Urban atmosphere conditions.
            
            Note: in very weird edge cases, a polluted atmosphere can actually result in slightly higher solar flux
            than clean air, due to increased back-scattering. For example, imagine it's near sunset, with the sun in
            the west, and your panel normal vector points east. Increased pollution can, in some edge cases,
            result in enough increased back-scattering (multipathing) that you have a smidge more illumination.

        albedo: The fraction of light that hits the ground that is reflected. Affects illumination from re-scattering
            when panels are tilted. Typical values for general terrestrial surfaces are 0.2, which is the default here.

            * Other values, taken from the Earthscan source (citation above):

                * Grass (July, August): 0.25
                * Lawn: 0.18 - 0.23
                * Dry grass: 0.28 - 0.32
                * Milled fields: 0.26
                * Barren soil: 0.17
                * Gravel: 0.18
                * Clean concrete: 0.30
                * Eroded concrete: 0.20
                * Clean cement: 0.55
                * Asphalt: 0.15
                * Forests: 0.05 - 0.18
                * Sandy areas: 0.10 - 0.25
                * Water: Strongly dependent on incidence angle; 0.05 - 0.22
                * Fresh snow: 0.80 - 0.90
                * Old snow: 0.45 - 0.70

    Returns: The solar power flux [W/m^2] on the panel.

        * Note: does not account for any potential reflectivity of the solar panel's coating itself.

    """
    flux_outside_atmosphere = solar_flux_outside_atmosphere_normal(day_of_year=day_of_year)

    solar_elevation = solar_elevation_angle(latitude, day_of_year, time)
    solar_azimuth = solar_azimuth_angle(latitude, day_of_year, time)

    relative_airmass = airmass(
        solar_elevation_angle=solar_elevation,
        altitude=altitude,
    )

    # Source: "Planning and installing..." Earthscan. Full citation in docstring above.
    if air_quality == 'typical':
        atmospheric_transmission_fraction = 0.70 ** (relative_airmass ** 0.678)
    elif air_quality == 'clean':
        atmospheric_transmission_fraction = 0.76 ** (relative_airmass ** 0.618)
    elif air_quality == 'polluted':
        atmospheric_transmission_fraction = 0.56 ** (relative_airmass ** 0.715)
    else:
        raise ValueError("Bad value of `air_quality`!")

    direct_normal_irradiance = np.where(
        solar_elevation > 0.,
        flux_outside_atmosphere * atmospheric_transmission_fraction,
        0.
    )

    absorption_and_scattering_losses = flux_outside_atmosphere - flux_outside_atmosphere * atmospheric_transmission_fraction

    scattering_losses = absorption_and_scattering_losses * (10. / 28.)
    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/air-mass
    # Indicates that absorption and scattering happen in a 18:10 ratio, at least in AM1.5 conditions. We extrapolate
    # this to all conditions.

    panel_tilt_angle = np.mod(panel_tilt_angle, 360)
    fraction_of_panel_facing_sky = np.where(
        panel_tilt_angle < 180,
        1 - panel_tilt_angle / 180,
        -1 + panel_tilt_angle / 180,
    )

    diffuse_irradiance = scattering_losses * atmospheric_transmission_fraction * (
            fraction_of_panel_facing_sky + albedo * (1 - fraction_of_panel_facing_sky)
    )
    # We assume that the in-scattering (i.e., diffuse irradiance) and the out-scattering (i.e., scattering losses in
    # the direct irradiance calculation) are equal, by argument of approximately parallel incident rays.
    # We further assume that any in-scattering must then once-again go through the absorption / re-scattering process,
    # which is identical to the original atmospheric transmission fraction.

    cosine_of_angle_between_panel_normal_and_sun = (
            np.cosd(solar_elevation) *
            np.sind(panel_tilt_angle) *
            np.cosd(panel_azimuth_angle - solar_azimuth)
            + np.sind(solar_elevation) * np.cosd(panel_tilt_angle)
    )
    cosine_of_angle_between_panel_normal_and_sun = np.fmax(
        cosine_of_angle_between_panel_normal_and_sun,
        0
    )  # Accounts for if you have a downwards-pointing panel while the sun is above you.

    # Source: https://www.pveducation.org/pvcdrom/properties-of-sunlight/arbitrary-orientation-and-tilt
    # Author of this code (Peter Sharpe) has manually verified correctness of this vector math.

    flux_on_panel = (
            direct_normal_irradiance * cosine_of_angle_between_panel_normal_and_sun
            + diffuse_irradiance
    )

    return flux_on_panel


def peak_sun_hours_per_day_on_horizontal(
        latitude: Union[float, np.ndarray],
        day_of_year: Union[int, float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    How many hours of equivalent peak sun do you get per day?
    :param latitude: Latitude [degrees]
    :param day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)
    :param time: Time since (local) solar noon [seconds]
    :return:
    """
    import warnings
    warnings.warn(
        "Use `solar_flux()` function from this module instead and integrate, which allows far greater granularity.",
        DeprecationWarning
    )
    time = np.linspace(0, 86400, 1000)
    fluxes = solar_flux(latitude, day_of_year, time)
    energy_per_area = np.sum(np.trapz(fluxes) * np.diff(time))

    duration_of_equivalent_peak_sun = energy_per_area / solar_flux(latitude, day_of_year, time=0.)

    sun_hours = duration_of_equivalent_peak_sun / 3600

    return sun_hours


def length_day(
        latitude: Union[float, np.ndarray],
        day_of_year: Union[int, float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Gives the duration where the sun is above the horizon on a given day.

    Args:

        latitude: Local geographic latitude [degrees]. Positive for north, negative for south.

        day_of_year: Julian day (1 == Jan. 1, 365 == Dec. 31)

    Returns: The duration where the sun is above the horizon on a given day. [seconds]

    """
    dec = declination_angle(day_of_year)

    constant = -np.sind(dec) * np.sind(latitude) / (np.cosd(dec) * np.cosd(latitude))

    sun_time_nondim = 2 * np.arccos(_prepare_for_inverse_trig(constant))
    sun_time = sun_time_nondim / (2 * np.pi) * 86400

    return sun_time


def mass_MPPT(
        power: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """
    Gives the estimated mass of a Maximum Power Point Tracking (MPPT) unit for solar energy
    collection. Based on regressions at AeroSandbox/studies/SolarMPPTMasses.

    Args:
        power: Power of MPPT [watts]

    Returns:
        Estimated MPPT mass [kg]
    """
    constant = 0.066343
    exponent = 0.515140
    return constant * power ** exponent


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    # plt.switch_backend('WebAgg')

    base_color = p.palettes['categorical'][0]
    quality_colors = {
        'clean'   : p.adjust_lightness(base_color, amount=1.2),
        'typical' : p.adjust_lightness(base_color, amount=0.7),
        'polluted': p.adjust_lightness(base_color, amount=0.2),
    }

    ##### Plot solar_flux() over the course of a day
    time = np.linspace(0, 86400, 86401)
    hour = time / 3600

    base_kwargs = dict(
        latitude=23.5,
        day_of_year=172,
        time=time,
    )

    fig, ax = plt.subplots(2, 1, figsize=(7, 6.5))
    plt.sca(ax[0])
    plt.title(f"Solar Flux on a Horizontal Surface Over A Day\n(Tropic of Cancer, Summer Solstice, Sea Level)")
    for q in quality_colors.keys():
        plt.plot(
            hour,
            solar_flux(
                **base_kwargs,
                air_quality=q
            ),
            color=quality_colors[q],
            label=f'ASB Model: {q.capitalize()} air'
        )

    plt.sca(ax[1])
    plt.title(f"Solar Flux on a Sun-Tracking Surface Over A Day\n(Tropic of Cancer, Summer Solstice, Sea Level)")
    for q in quality_colors.keys():
        plt.plot(
            hour,
            solar_flux(
                **base_kwargs,
                panel_tilt_angle=90 - solar_elevation_angle(**base_kwargs),
                panel_azimuth_angle=solar_azimuth_angle(**base_kwargs),
                air_quality=q
            ),
            color=quality_colors[q],
            label=f'ASB Model: {q.capitalize()} air'
        )

    for a in ax:
        plt.sca(a)
        plt.xlabel("Time after Local Solar Noon [hours]")
        plt.ylabel("Solar Flux [$W/m^2$]")
        plt.xlim(0, 24)
        plt.ylim(-10, 1200)
        p.set_ticks(3, 0.5, 200, 50)

    plt.sca(ax[0])
    p.show_plot()

    ##### Plot solar_flux() as a function of elevation angle, and compare to data to validate.
    # Source: Ed. (2008), Table 1.1, Earthscan with the International Institute for Environment and Development, Deutsche Gesellschaft für Sonnenenergie. ISBN 1-84407-442-0.
    # Via: https://en.wikipedia.org/wiki/Air_mass_(solar_energy)#Solar_intensity

    # Values here give lower and upper bounds for measured solar flux on a typical clear day, varying primarily due
    # to pollution.
    raw_data = """\
z [deg],AM [-],Solar Flux Lower Bound [W/m^2],Solar Flux Upper Bound [W/m^2]
0,1,840,1130
23,1.09,800,1110
30,1.15,780,1100
45,1.41,710,1060
48.2,1.5,680,1050
60,2,560,970
70,2.9,430,880
75,3.8,330,800
80,5.6,200,660
85,10,85,480
90,38,6,34
"""
    import pandas as pd
    from io import StringIO

    delimiter = "\t"
    df = pd.read_csv(
        StringIO(raw_data),
        delimiter=','
    )
    df["Solar Flux [W/m^2]"] = (df['Solar Flux Lower Bound [W/m^2]'] + df['Solar Flux Upper Bound [W/m^2]']) / 2

    fluxes = solar_flux(
        **base_kwargs,
        panel_tilt_angle=90 - solar_elevation_angle(**base_kwargs),
        panel_azimuth_angle=solar_azimuth_angle(**base_kwargs),
    )
    elevations = solar_elevation_angle(
        **base_kwargs
    )

    fig, ax = plt.subplots()
    for q in quality_colors.keys():
        plt.plot(
            solar_elevation_angle(**base_kwargs),
            solar_flux(
                **base_kwargs,
                panel_tilt_angle=90 - solar_elevation_angle(**base_kwargs),
                panel_azimuth_angle=solar_azimuth_angle(**base_kwargs),
                air_quality=q
            ),
            color=quality_colors[q],
            label=f'ASB Model: {q.capitalize()} air',
            zorder=3
        )

    data_color = p.palettes['categorical'][1]

    plt.fill_between(
        x=90 - df['z [deg]'].values,
        y1=df['Solar Flux Lower Bound [W/m^2]'],
        y2=df['Solar Flux Upper Bound [W/m^2]'],
        color=data_color,
        alpha=0.4,
        label='Experimental Data Range\n(due to Pollution)',
        zorder=2.9,
    )
    for d in ['Lower', 'Upper']:
        plt.plot(
            90 - df['z [deg]'].values,
            df[f'Solar Flux {d} Bound [W/m^2]'],
            ".",
            color=data_color,
            alpha=0.7,
            zorder=2.95
        )

    plt.annotate(
        text='Data: "Planning and Installing Photovoltaic Systems".\nEarthscan (2008), ISBN 1-84407-442-0.',
        xy=(0.02, 0.98),
        xycoords="axes fraction",
        ha="left",
        va='top',
        fontsize=9
    )
    plt.xlim(-5, 90)
    p.set_ticks(15, 5, 200, 50)

    p.show_plot(
        f"Sun Position vs. Solar Flux on a Sun-Tracking Surface",
        f"Solar Elevation Angle [deg]",
        "Solar Flux [$W/m^2$]"
    )



import aerosandbox.numpy as np
import pytest
import casadi as cas
import numpy as onp


def test_interp():
    x_np = np.arange(5)
    y_np = x_np + 10

    for x, y in zip(
            [x_np, cas.DM(x_np)],
            [y_np, cas.DM(y_np)]
    ):

        assert np.interp(0, x, y) == pytest.approx(10)
        assert np.interp(4, x, y) == pytest.approx(14)
        assert np.interp(0.5, x, y) == pytest.approx(10.5)
        # Extrapolation has different behaviours between casadi and numpy. TODO investigate why
        # assert np.interp(-1, x, y) == pytest.approx(10)
        # assert np.interp(5, x, y) == pytest.approx(14)
        assert np.interp(-1, x, y, left=-10) == pytest.approx(-10)
        assert np.interp(5, x, y, right=-10) == pytest.approx(-10)
        assert np.interp(5, x, y, period=4) == pytest.approx(11)


def test_interpn_linear():
    ### NumPy test

    def value_func_3d(x, y, z):
        return 2 * x + 3 * y - z

    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 20)
    z = np.linspace(0, 5, 30)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))

    point = np.array([2.21, 3.12, 1.15])
    value = np.interpn(
        points, values, point
    )
    assert value == pytest.approx(value_func_3d(*point))

    ### CasADi test
    point = cas.DM(point)
    value = np.interpn(
        points, values, point
    )
    assert value == pytest.approx(float(value_func_3d(point[0], point[1], point[2])))


def test_interpn_linear_multiple_samples():
    ### NumPy test

    def value_func_3d(x, y, z):
        return 2 * x + 3 * y - z

    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 20)
    z = np.linspace(0, 5, 30)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))

    point = np.array([
        [2.21, 3.12, 1.15],
        [3.42, 0.81, 2.43]
    ])
    value = np.interpn(
        points, values, point
    )
    assert np.all(
        value == pytest.approx(
            value_func_3d(
                *[
                    point[:, i] for i in range(point.shape[1])
                ]
            )
        )
    )
    assert len(value) == 2

    ### CasADi test
    point = cas.DM(point)
    value = np.interpn(
        points, values, point
    )
    value_actual = value_func_3d(
        *[
            np.array(point[:, i]) for i in range(point.shape[1])
        ]
    )
    for i in range(len(value)):
        assert value[i] == pytest.approx(float(value_actual[i]))
    assert value.shape == (2,)


def test_interpn_bspline_casadi():
    """
    The bspline method should interpolate seperable cubic multidimensional polynomials exactly.
    """

    def func(x, y, z):  # Sphere function
        return x ** 3 + y ** 3 + z ** 3

    x = np.linspace(-5, 5, 10)
    y = np.linspace(-5, 5, 20)
    z = np.linspace(-5, 5, 30)
    points = (x, y, z)
    values = func(
        *np.meshgrid(*points, indexing="ij")
    )

    point = np.array([0.4, 0.5, 0.6])
    value = np.interpn(
        points, values, point, method="bspline"
    )

    assert value == pytest.approx(func(*point))


def test_interpn_bounds_error_one_sample():
    def value_func_3d(x, y, z):
        return 2 * x + 3 * y - z

    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 20)
    z = np.linspace(0, 5, 30)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))

    point = np.array([5.21, 3.12, 1.15])
    with pytest.raises(ValueError):
        value = np.interpn(
            points, values, point
        )

    ### CasADi test
    point = cas.DM(point)
    with pytest.raises(ValueError):
        value = np.interpn(
            points, values, point
        )


def test_interpn_bounds_error_multiple_samples():
    def value_func_3d(x, y, z):
        return 2 * x + 3 * y - z

    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 20)
    z = np.linspace(0, 5, 30)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))

    point = np.array([
        [2.21, 3.12, 1.15],
        [3.42, 5.81, 2.43]
    ])
    with pytest.raises(ValueError):
        value = np.interpn(
            points, values, point
        )

    ### CasADi test
    point = cas.DM(point)
    with pytest.raises(ValueError):
        value = np.interpn(
            points, values, point
        )


def test_interpn_fill_value():
    def value_func_3d(x, y, z):
        return 2 * x + 3 * y - z

    x = np.linspace(0, 5, 10)
    y = np.linspace(0, 5, 20)
    z = np.linspace(0, 5, 30)
    points = (x, y, z)
    values = value_func_3d(*np.meshgrid(*points, indexing="ij"))

    point = np.array([5.21, 3.12, 1.15])

    value = np.interpn(
        points, values, point,
        method="bspline",
        bounds_error=False,
        fill_value=-17
    )
    assert value == pytest.approx(-17)

    value = np.interpn(
        points, values, point,
        method="bspline",
        bounds_error=False,
    )
    assert np.isnan(value)

    value = np.interpn(
        points, values, point,
        method="bspline",
        bounds_error=None,
        fill_value=None
    )
    assert value == pytest.approx(value_func_3d(5, 3.12, 1.15))


if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np


def Cd_cylinder(
        Re_D: float,
        mach: float = 0.,
        include_mach_effects=True,
        subcritical_only=False
) -> float:
    """
    Returns the drag coefficient of a cylinder in crossflow as a function of its Reynolds number and Mach.

    Args:
        Re_D: Reynolds number, referenced to diameter
        mach: Mach number
        include_mach_effects: If this is set False, it assumes Mach = 0, which simplifies the computation.
        subcritical_only: Determines whether the model models purely subcritical (Re < 300k) cylinder flows. Useful, since
    this model is now convex and can be more well-behaved.

    Returns:

    # TODO rework this function to use tanh blending, which will mitigate overflows

    """

    ##### Do the viscous part of the computation
    csigc = 5.5766722118597247
    csigh = 23.7460859935990563
    csub0 = -0.6989492360435040
    csub1 = 1.0465189382830078
    csub2 = 0.7044228755898569
    csub3 = 0.0846501115443938
    csup0 = -0.0823564417206403
    csupc = 6.8020230357616764
    csuph = 9.9999999999999787
    csupscl = -0.4570690347113859

    x = np.log10(np.abs(Re_D) + 1e-16)

    if subcritical_only:
        Cd_mach_0 = 10 ** (csub0 * x + csub1) + csub2 + csub3 * x
    else:
        log10_Cd = (
                (np.log10(10 ** (csub0 * x + csub1) + csub2 + csub3 * x))
                * (1 - 1 / (1 + np.exp(-csigh * (x - csigc))))
                + (csup0 + csupscl / csuph * np.log(np.exp(csuph * (csupc - x)) + 1))
                * (1 / (1 + np.exp(-csigh * (x - csigc))))
        )
        Cd_mach_0 = 10 ** log10_Cd

    ##### Do the compressible part of the computation
    if include_mach_effects:
        m = mach
        p = {'a_sub'    : 0.03458900259594298,
             'a_sup'    : -0.7129528087049688,
             'cd_sub'   : 1.163206940186374,
             'cd_sup'   : 1.2899213533122527,
             's_sub'    : 3.436601777569716,
             's_sup'    : -1.37123096976983,
             'trans'    : 1.022819211244295,
             'trans_str': 19.017600596069848}

        Cd_over_Cd_mach_0 = np.blend(
            p["trans_str"] * (m - p["trans"]),
            p["cd_sup"] + np.exp(p["a_sup"] + p["s_sup"] * (m - p["trans"])),
            p["cd_sub"] + np.exp(p["a_sub"] + p["s_sub"] * (m - p["trans"]))
        ) / 1.1940010047391572

        Cd = Cd_mach_0 * Cd_over_Cd_mach_0

    else:
        Cd = Cd_mach_0

    return Cd


def Cf_flat_plate(
        Re_L: float,
        method="hybrid-sharpe-convex"
) -> float:
    """
    Returns the mean skin friction coefficient over a flat plate.

    Don't forget to double it (two sides) if you want a drag coefficient.

    Args:

        Re_L: Reynolds number, normalized to the length of the flat plate.

        method: The method of computing the skin friction coefficient. One of:

            * "blasius": Uses the Blasius solution. Citing Cengel and Cimbala, "Fluid Mechanics: Fundamentals and
            Applications", Table 10-4.

                Valid approximately for Re_L <= 5e5.

            * "turbulent": Uses turbulent correlations for smooth plates. Citing Cengel and Cimbala,
            "Fluid Mechanics: Fundamentals and Applications", Table 10-4.

                Valid approximately for 5e5 <= Re_L <= 1e7.

            * "hybrid-cengel": Uses turbulent correlations for smooth plates, but accounts for a
            non-negligible laminar run at the beginning of the plate. Citing Cengel and Cimbala, "Fluid Mechanics:
            Fundamentals and Applications", Table 10-4. Returns: Mean skin friction coefficient over a flat plate.

                Valid approximately for 5e5 <= Re_L <= 1e7.

            * "hybrid-schlichting": Schlichting's model, that roughly accounts for a non-negligtible laminar run.
            Citing "Boundary Layer Theory" 7th Ed., pg. 644

            * "hybrid-sharpe-convex": A hybrid model that blends the Blasius and Schlichting models. Convex in
            log-log space; however, it may overlook some truly nonconvex behavior near transitional Reynolds numbers.

            * "hybrid-sharpe-nonconvex": A hybrid model that blends the Blasius and Cengel models. Nonconvex in
            log-log-space; however, it may capture some truly nonconvex behavior near transitional Reynolds numbers.

    Returns:

        C_f: The skin friction coefficient, normalized to the length of the plate.

    You can view all of these functions graphically using
    `aerosandbox.library.aerodynamics.test_aerodynamics.test_Cf_flat_plate.py`

    """
    Re_L = np.abs(Re_L)

    if method == "blasius":
        return 1.328 / Re_L ** 0.5
    elif method == "turbulent":
        return 0.074 / Re_L ** (1 / 5)
    elif method == "hybrid-cengel":
        return 0.074 / Re_L ** (1 / 5) - 1742 / Re_L
    elif method == "hybrid-schlichting":
        return 0.02666 * Re_L ** -0.139
    elif method == "hybrid-sharpe-convex":
        return np.softmax(
            Cf_flat_plate(Re_L, method="blasius"),
            Cf_flat_plate(Re_L, method="hybrid-schlichting"),
            hardness=1e3
        )
    elif method == "hybrid-sharpe-nonconvex":
        return np.softmax(
            Cf_flat_plate(Re_L, method="blasius"),
            Cf_flat_plate(Re_L, method="hybrid-cengel"),
            hardness=1e3
        )


def Cl_flat_plate(alpha, Re_c=None):
    """
    Returns the approximate lift coefficient of a flat plate, following thin airfoil theory.
    :param alpha: Angle of attack [deg]
    :param Re_c: Reynolds number, normalized to the length of the flat plate.
    :return: Approximate lift coefficient.
    """
    if Re_c is not None:
        from warnings import warn
        warn("`Re_c` input will be deprecated in a future version.")

    alpha_rad = alpha * np.pi / 180
    return 2 * np.pi * alpha_rad


def Cd_flat_plate_normal():
    """
    Returns the drag coefficient of a flat plat oriented normal to the flow (i.e., alpha = 90 deg).

    Uses results from Tian, Xinliang, Muk Chen Ong, Jianmin Yang, and Dag Myrhaug. “Large-Eddy Simulation of the Flow
    Normal to a Flat Plate Including Corner Effects at a High Reynolds Number.” Journal of Fluids and Structures 49 (
    August 2014): 149–69. https://doi.org/10.1016/j.jfluidstructs.2014.04.008.

    Note: Cd for this case is effectively invariant of Re.

    Returns: Drag coefficient

    """
    return 2.202


import warnings


def Cl_2412(alpha, Re_c):
    # A curve fit I did to a NACA 2412 airfoil, 2D XFoil data
    # Within -2 < alpha < 12 and 10^5 < Re_c < 10^7, has R^2 = 0.9892

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    return 0.2568 + 0.1206 * alpha - 0.002018 * alpha ** 2


def Cd_profile_2412(alpha, Re_c):
    # A curve fit I did to a NACA 2412 airfoil in incompressible flow.
    # Within -2 < alpha < 12 and 10^5 < Re_c < 10^7, has R^2 = 0.9713

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    Re_c = np.maximum(Re_c, 1)
    log_Re = np.log(Re_c)

    CD0 = -5.249
    Re0 = 15.61
    Re1 = 15.31
    alpha0 = 1.049
    alpha1 = -4.715
    cx = 0.009528
    cxy = -0.00588
    cy = 0.04838

    log_CD = CD0 + cx * (alpha - alpha0) ** 2 + cy * (log_Re - Re0) ** 2 + cxy * (alpha - alpha1) * (
            log_Re - Re1)  # basically, a rotated paraboloid in logspace
    CD = np.exp(log_CD)

    return CD


def Cl_e216(alpha, Re_c):
    # A curve fit I did to a Eppler 216 (e216) airfoil, 2D XFoil data. Incompressible flow.
    # Within -2 < alpha < 12 and 10^4 < Re_c < 10^6, has R^2 = 0.9994
    # Likely valid from -6 < alpha < 12 and 10^4 < Re_c < Inf.
    # See: C:\Projects\GitHub\firefly_aerodynamics\Gists and Ideas\XFoil Drag Fitting\e216

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    Re_c = np.fmax(Re_c, 1)
    log10_Re = np.log10(Re_c)

    # Coeffs
    a1l = 3.0904412662858878e-02
    a1t = 9.6452654383488254e-02
    a4t = -2.5633334023068302e-05
    asl = 6.4175433185427011e-01
    atr = 3.6775107602844948e-01
    c0l = -2.5909363461176749e-01
    c0t = 8.3824440586718862e-01
    ctr = 1.1431810545735890e+02
    ksl = 5.3416670116733611e-01
    rtr = 3.9713338634462829e+01
    rtr2 = -3.3634858542657771e+00
    xsl = -1.2220899840236835e-01

    a = alpha
    r = log10_Re

    Cl = (c0t + a1t * a + a4t * a ** 4) * 1 / (1 + np.exp(ctr - rtr * r - atr * a - rtr2 * r ** 2)) + (
            c0l + a1l * a + asl / (1 + np.exp(-ksl * (a - xsl)))) * (
                 1 - 1 / (1 + np.exp(ctr - rtr * r - atr * a - rtr2 * r ** 2)))

    return Cl


def Cd_profile_e216(alpha, Re_c):
    # A curve fit I did to a Eppler 216 (e216) airfoil, 2D XFoil data. Incompressible flow.
    # Within -2 < alpha < 12 and 10^4 < Re_c < 10^6, has R^2 = 0.9995
    # Likely valid from -6 < alpha < 12 and 10^4 < Re_c < 10^6.
    # see: C:\Projects\GitHub\firefly_aerodynamics\Gists and Ideas\XFoil Drag Fitting\e216

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    Re_c = np.fmax(Re_c, 1)
    log10_Re = np.log10(Re_c)

    # Coeffs
    a1l = 4.7167470806940448e-02
    a1t = 7.5663005080888857e-02
    a2l = 8.7552076545610764e-04
    a4t = 1.1220763679805319e-05
    atr = 4.2456038382581129e-01
    c0l = -1.4099657419753771e+00
    c0t = -2.3855286371940609e+00
    ctr = 9.1474872611212135e+01
    rtr = 3.0218483612170434e+01
    rtr2 = -2.4515094313899279e+00

    a = alpha
    r = log10_Re

    log10_Cd = (c0t + a1t * a + a4t * a ** 4) * 1 / (1 + np.exp(ctr - rtr * r - atr * a - rtr2 * r ** 2)) + (
            c0l + a1l * a + a2l * a ** 2) * (1 - 1 / (1 + np.exp(ctr - rtr * r - atr * a - rtr2 * r ** 2)))

    Cd = 10 ** log10_Cd

    return Cd


def Cd_wave_e216(Cl, mach, sweep=0.):
    r"""
    A curve fit I did to Eppler 216 airfoil data.
    Within -0.4 < CL < 0.75 and 0 < mach < ~0.9, has R^2 = 0.9982.
    See: C:\Projects\GitHub\firefly_aerodynamics\MSES Interface\analysis\e216
    :param Cl: Lift coefficient
    :param mach: Mach number
    :param sweep: Sweep angle, in deg
    :return: Wave drag coefficient.
    """

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    mach = np.fmax(mach, 0)
    mach_perpendicular = mach * np.cosd(sweep)  # Relation from FVA Eq. 8.176
    Cl_perpendicular = Cl / np.cosd(sweep) ** 2  # Relation from FVA Eq. 8.177

    # Coeffs
    c0 = 7.2685945744797997e-01
    c1 = -1.5483144040727698e-01
    c3 = 2.1305118052118968e-01
    c4 = 7.8812272501525316e-01
    c5 = 3.3888938102072169e-03
    l0 = 1.5298928303149546e+00
    l1 = 5.2389999717540392e-01

    m = mach_perpendicular
    l = Cl_perpendicular

    Cd_wave = (np.fmax(m - (c0 + c1 * np.sqrt(c3 + (l - c4) ** 2) + c5 * l), 0) * (l0 + l1 * l)) ** 2

    return Cd_wave


def Cl_rae2822(alpha, Re_c):
    # A curve fit I did to a RAE2822 airfoil, 2D XFoil data. Incompressible flow.
    # Within -2 < alpha < 12 and 10^4 < Re_c < 10^6, has R^2 = 0.9857
    # Likely valid from -6 < alpha < 12 and 10^4 < Re_c < 10^6.
    # See: C:\Projects\GitHub\firefly_aerodynamics\Gists and Ideas\XFoil Drag Fitting\rae2822

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    Re_c = np.fmax(Re_c, 1)
    log10_Re = np.log10(Re_c)

    # Coeffs
    a1l = 5.5686866813855172e-02
    a1t = 9.7472055628494134e-02
    a4l = -7.2145733312046152e-09
    a4t = -3.6886704372829236e-06
    atr = 8.3723547264375520e-01
    atr2 = -8.3128119739031697e-02
    c0l = -4.9103908291438701e-02
    c0t = 2.3903424824298553e-01
    ctr = 1.3082854754897108e+01
    rtr = 2.6963082864300731e+00

    a = alpha
    r = log10_Re

    Cl = (c0t + a1t * a + a4t * a ** 4) * 1 / (1 + np.exp(ctr - rtr * r - atr * a - atr2 * a ** 2)) + (
            c0l + a1l * a + a4l * a ** 4) * (1 - 1 / (1 + np.exp(ctr - rtr * r - atr * a - atr2 * a ** 2)))

    return Cl


def Cd_profile_rae2822(alpha, Re_c):
    # A curve fit I did to a RAE2822 airfoil, 2D XFoil data. Incompressible flow.
    # Within -2 < alpha < 12 and 10^4 < Re_c < 10^6, has R^2 = 0.9995
    # Likely valid from -6 < alpha < 12 and 10^4 < Re_c < Inf.
    # see: C:\Projects\GitHub\firefly_aerodynamics\Gists and Ideas\XFoil Drag Fitting\e216

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    Re_c = np.fmax(Re_c, 1)
    log10_Re = np.log10(Re_c)

    # Coeffs
    at = 8.1034027621509015e+00
    c0l = -8.4296746456429639e-01
    c0t = -1.3700609138855402e+00
    kart = -4.1609994062600880e-01
    kat = 5.9510959342452441e-01
    krt = -7.1938030052506197e-01
    r1l = 1.1548628822014631e-01
    r1t = -4.9133662875044504e-01
    rt = 5.0070459892411696e+00

    a = alpha
    r = log10_Re

    log10_Cd = (c0t + r1t * (r - 4)) * (
            1 / (1 + np.exp(kat * (a - at) + krt * (r - rt) + kart * (a - at) * (r - rt)))) + (
                       c0l + r1l * (r - 4)) * (
                       1 - 1 / (1 + np.exp(kat * (a - at) + krt * (r - rt) + kart * (a - at) * (r - rt))))

    Cd = 10 ** log10_Cd

    return Cd


def Cd_wave_rae2822(Cl, mach, sweep=0.):
    r"""
    A curve fit I did to RAE2822 airfoil data.
    Within -0.4 < CL < 0.75 and 0 < mach < ~0.9, has R^2 = 0.9982.
    See: C:\Projects\GitHub\firefly_aerodynamics\MSES Interface\analysis\rae2822
    :param Cl: Lift coefficient
    :param mach: Mach number
    :param sweep: Sweep angle, in deg
    :return: Wave drag coefficient.
    """

    warnings.warn(
        "This function is deprecated. Use `asb.Airfoil.get_aero_from_neuralfoil()` instead.",
        DeprecationWarning,
    )
    mach = np.fmax(mach, 0)
    mach_perpendicular = mach * np.cosd(sweep)  # Relation from FVA Eq. 8.176
    Cl_perpendicular = Cl / np.cosd(sweep) ** 2  # Relation from FVA Eq. 8.177

    # Coeffs
    c2 = 4.5776476424519119e+00
    mc0 = 9.5623337929607111e-01
    mc1 = 2.0552787101770234e-01
    mc2 = 1.1259268018737063e+00
    mc3 = 1.9538856688443659e-01

    m = mach_perpendicular
    l = Cl_perpendicular

    Cd_wave = np.fmax(m - (mc0 - mc1 * np.sqrt(mc2 + (l - mc3) ** 2)), 0) ** 2 * c2

    return Cd_wave


def fuselage_upsweep_drag_area(
        upsweep_angle_rad: float,
        fuselage_xsec_area_max: float,
) -> float:
    """
    Calculates the drag area (in m^2) of the aft end of a fuselage with a given upsweep angle.

    Upsweep is the characteristic shape seen on the aft end of many fuselages in transport aircraft, where the
    centerline of the fuselage is angled upwards near the aft end. This is done to reduce the required landing gear
    height for adequate takeoff rotation, which in turn reduces mass. This nonzero centerline angle can cause some
    separation drag, which is predicted here.

    Equation is from Raymer, Aircraft Design: A Conceptual Approach, 5th Ed., Eq. 12.36, pg. 440.

    Args:
        upsweep_angle_rad: The upsweep angle of the aft end of the fuselage relative to the centerline, in radians.

        fuselage_xsec_area_max: The maximum cross-sectional area of the fuselage, in m^2.

    Returns: The drag area of the aft end of the fuselage [m^2]. This is equivalent to D/q, where D is the drag force
    and q is the dynamic pressure.
    """
    return 3.83 * np.abs(upsweep_angle_rad) ** 2.5 * fuselage_xsec_area_max


if __name__ == "__main__":
    pass
    # # Run some checks
    # import matplotlib.pyplot as plt
    # import matplotlib.style as style
    # import plotly.express as px
    # from plotly import io
    #
    # io.renderers.default = "browser"
    #
    # # # E216 checks
    # alpha_inputs = np.linspace(-6, 12, 200)
    # Re_inputs = np.logspace(4, 6, 200)
    # alphas = []
    # Res = []
    # CLs = []
    # CDs = []
    # for alpha in alpha_inputs:
    #     for Re in Re_inputs:
    #         alphas.append(alpha)
    #         Res.append(Re)
    #         CLs.append(Cl_e216(alpha, Re))
    #         CDs.append(Cd_profile_e216(alpha, Re))
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=CLs,
    #     size=np.ones_like(alphas),
    #     color=CLs,
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CL"}
    # ).show()
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=CDs,
    #     size=np.ones_like(alphas),
    #     color=CDs,
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CD"}
    # ).show()
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=np.array(CLs) / np.array(CDs),
    #     size=np.ones_like(alphas),
    #     color=np.array(CLs) / np.array(CDs),
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CL/CD"}
    # ).show()
    #
    # # # rae2822 checks
    # alpha_inputs = np.linspace(-6, 12)
    # Re_inputs = np.logspace(4, 6)
    # alphas = []
    # Res = []
    # CLs = []
    # CDs = []
    # for alpha in alpha_inputs:
    #     for Re in Re_inputs:
    #         alphas.append(alpha)
    #         Res.append(Re)
    #         CLs.append(Cl_rae2822(alpha, Re))
    #         CDs.append(Cd_profile_rae2822(alpha, Re))
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=CLs,
    #     size=np.ones_like(alphas),
    #     color=CLs,
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CL"}
    # ).show()
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=CDs,
    #     size=np.ones_like(alphas),
    #     color=CDs,
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CD"}
    # ).show()
    # px.scatter_3d(
    #     x=alphas,
    #     y=Res,
    #     z=np.array(CLs) / np.array(CDs),
    #     size=np.ones_like(alphas),
    #     color=np.array(CLs) / np.array(CDs),
    #     log_y=True,
    #     labels={"x": "alphas", "y": "Re", "z": "CL/CD"}
    # ).show()
    #
    # # Cd_wave_e216 check
    # CL_inputs = np.linspace(-0.4, 1)
    # mach_inputs = np.linspace(0.3, 1)
    # CLs = []
    # machs = []
    # CD_waves = []
    # for CL in CL_inputs:
    #     for mach in mach_inputs:
    #         CLs.append(CL)
    #         machs.append(mach)
    #         CD_waves.append(Cd_wave_e216(CL, mach))
    # px.scatter_3d(
    #     x=CLs,
    #     y=machs,
    #     z=CD_waves,
    #     size=np.ones_like(CD_waves),
    #     color=CD_waves,
    #     title="E216 Fit",
    #     labels={"x": "CL", "y": "Mach", "z": "CD_wave"},
    #     range_z=(0, 200e-4)
    # ).show()
    #
    # # Cd_wave_rae2822 check
    # CL_inputs = np.linspace(-0.4, 1)
    # mach_inputs = np.linspace(0.3, 1)
    # CLs = []
    # machs = []
    # CD_waves = []
    # for CL in CL_inputs:
    #     for mach in mach_inputs:
    #         CLs.append(CL)
    #         machs.append(mach)
    #         CD_waves.append(Cd_wave_rae2822(CL, mach))
    # px.scatter_3d(
    #     x=CLs,
    #     y=machs,
    #     z=CD_waves,
    #     size=np.ones_like(CD_waves),
    #     color=CD_waves,
    #     title="RAE2822 Fit",
    #     labels={"x": "CL", "y": "Mach", "z": "CD_wave"},
    #     # range_z=(0, 200e-4)
    # ).show()
    #
    # # Cd_wave_Korn check
    # CL_inputs = np.linspace(-0.4, 1)
    # mach_inputs = np.linspace(0.3, 1)
    # CLs = []
    # machs = []
    # CD_waves = []
    # for CL in CL_inputs:
    #     for mach in mach_inputs:
    #         CLs.append(CL)
    #         machs.append(mach)
    #         CD_waves.append(float(Cd_wave_Korn(CL, t_over_c=0.121, mach=mach, kappa_A=0.95)))
    # px.scatter_3d(
    #     x=CLs,
    #     y=machs,
    #     z=CD_waves,
    #     size=list(np.ones_like(CD_waves)),
    #     color=CD_waves,
    #     title="Korn Equation",
    #     labels={"x": "CL", "y": "Mach", "z": "CD_wave"},
    #     range_z=(0, 200e-4)
    # ).show()
    #
    # # # Cylinder Drag Check
    # Res = np.logspace(-1, 8, 300)
    # CDs = Cd_cylinder(Res)
    # CDs_s = Cd_cylinder(Res, True)
    #
    # plt.loglog(Res, CDs, label="Full Model")
    # plt.loglog(Res, CDs_s, label="Subcrit. Only Model")
    # plt.xlabel("Re")
    # plt.ylabel("CD")
    # plt.title("Cylinder Drag Checking")
    # plt.legend()
    # plt.show()



import numpy as _onp
import casadi as _cas
from aerosandbox.numpy.determine_type import is_casadi_type


def linspace(
        start: float = 0.,
        stop: float = 1.,
        num: int = 50
):
    """
    Returns evenly spaced numbers over a specified interval.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.linspace.html
    """
    if not is_casadi_type([start, stop, num], recursive=True):
        return _onp.linspace(start, stop, num)
    else:
        return _cas.linspace(start, stop, num)


def cosspace(
        start: float = 0.,
        stop: float = 1.,
        num: int = 50
):
    """
    Makes a cosine-spaced vector.

    Cosine spacing is useful because these correspond to Chebyshev nodes: https://en.wikipedia.org/wiki/Chebyshev_nodes

    To learn more about cosine spacing, see this: https://youtu.be/VSvsVgGbN7I

    Args:
        start: Value to start at.
        end: Value to end at.
        num: Number of points in the vector.
    """
    mean = (stop + start) / 2
    amp = (stop - start) / 2
    ones = 0 * start + 1
    spaced_array = mean + amp * _onp.cos(
        linspace(
            _onp.pi * ones,
            0 * ones,
            num
        )
    )

    # Fix the endpoints, which might not be exactly right due to floating-point error.
    spaced_array[0] = start
    spaced_array[-1] = stop

    return spaced_array


def sinspace(
        start: float = 0.,
        stop: float = 1.,
        num: int = 50,
        reverse_spacing: bool = False,
):
    """
    Makes a sine-spaced vector. By default, bunches points near the start.

    Sine spacing is exactly identical to half of a cosine-spaced distrubution, in terms of relative separations.

    To learn more about sine spacing and cosine spacing, see this: https://youtu.be/VSvsVgGbN7I

    Args:

        start: Value to start at.

        end: Value to end at.

        num: Number of points in the vector.

        reverse_spacing: Does negative-sine spacing. In other words, if this is True, the points will be bunched near
        the `stop` rather than at the `start`.

    Points are bunched up near the `start` of the interval by default. To reverse this, use the `reverse_spacing`
    parameter.
    """
    if reverse_spacing:
        return sinspace(stop, start, num)[::-1]
    ones = 0 * start + 1
    spaced_array = (
            start + (stop - start) * (1 - _onp.cos(linspace(
        0 * ones,
        _onp.pi / 2 * ones,
        num
    ))
                                      )
    )
    # Fix the endpoints, which might not be exactly right due to floating-point error.
    spaced_array[0] = start
    spaced_array[-1] = stop

    return spaced_array


def logspace(
        start: float = 0.,
        stop: float = 1.,
        num: int = 50
):
    """
    Return numbers spaced evenly on a log scale.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.logspace.html
    """
    if not is_casadi_type([start, stop, num], recursive=True):
        return _onp.logspace(start, stop, num)
    else:
        return 10 ** linspace(start, stop, num)


def geomspace(
        start: float = 1.,
        stop: float = 10.,
        num: int = 50
):
    """
    Return numbers spaced evenly on a log scale (a geometric progression).

    This is similar to logspace, but with endpoints specified directly.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.geomspace.html
    """
    if not is_casadi_type([start, stop, num], recursive=True):
        return _onp.geomspace(start, stop, num)
    else:
        if start <= 0 or stop <= 0:
            raise ValueError("Both start and stop must be positive!")
        spaced_array = _onp.log10(10 ** linspace(start, stop, num))

        # Fix the endpoints, which might not be exactly right due to floating-point error.
        spaced_array[0] = start
        spaced_array[-1] = stop

        return spaced_array



"""
This script generates data used to estimate spar mass as a function of lift force (i.e. total weight) and span.
"""

### Imports
from aerosandbox.structures.legacy.beams import *
import scipy.io as sio
import copy

### Set up sweep variables
masses = np.linspace(50, 800, 50)
spans = np.linspace(30, 90, 50)

Masses, Spans = np.meshgrid(masses, spans, indexing="ij")
Spar_Masses = np.zeros_like(Masses)

### Set up problem
opti = cas.Opti()
mass = opti.parameter()
span = opti.parameter()
beam = TubeBeam1(
    opti=opti,
    length=span / 2,
    points_per_point_load=50,
    diameter_guess=100,
    bending=True,
    torsion=False
)
lift_force = 9.81 * mass
beam.add_point_load(
    location=span / 2 / 2 + 1,
    force=-lift_force / 2
)
beam.add_uniform_load(force=lift_force / 2)
beam.setup()

# Tip deflection constraint
opti.subject_to([
    # beam.u[-1] < 2,  # Source: http://web.mit.edu/drela/Public/web/hpa/hpa_structure.pdf
    # beam.u[-1] > -2  # Source: http://web.mit.edu/drela/Public/web/hpa/hpa_structure.pdf
    beam.du * 180 / cas.pi < 10,
    beam.du * 180 / cas.pi > -10
])
opti.subject_to([
    cas.diff(cas.diff(beam.nominal_diameter)) < 0.002,
    cas.diff(cas.diff(beam.nominal_diameter)) > -0.002,
])

opti.minimize(beam.mass)

p_opts = {}
s_opts = {}
s_opts["max_iter"] = 1000  # If you need to interrupt, just use ctrl+c
# s_opts["mu_strategy"] = "adaptive"
# s_opts["watchdog_shortened_iter_trigger"] = 1
# s_opts["expect_infeasible_problem"]="yes"
# s_opts["start_with_resto"] = "yes"
# s_opts["required_infeasibility_reduction"] = 0.001
opti.solver('ipopt', p_opts, s_opts)

### Do the sweep
for i in range(len(masses)):
    iterable = range(len(spans))
    iterable = iterable[::-1] if i % 2 != 0 else iterable
    for j in iterable:
        opti.set_value(mass, Masses[i, j])
        opti.set_value(span, Spans[i, j])

        sol = opti.solve()
        opti.set_initial(sol.value_variables())
        beam_sol = sol(beam)

        Spar_Masses[i, j] = beam_sol.mass

sio.savemat("data_double_boom.mat",
            {
                "Masses"     : Masses,
                "Spans"      : Spans,
                "Spar_Masses": Spar_Masses
            })



from math import pi

# Lengths (equivalents in meters)
meter = 1
inch = 0.0254
foot = 12 * inch
yard = 3 * foot
mile = 5280 * foot
naut_mile = 1852

# Masses (equivalents in kilograms)
kg = 1
gram = 0.001
slug = 14.59390
lbm = 0.45359237
pound = lbm
short_ton = 2000 * lbm
long_ton = 2240 * lbm
oz = lbm / 16

# Time (equivalents in seconds)
second = 1
minute = 60
hour = 60 * minute
day = 24 * hour
week = 7 * day
year = 365.2425 * day
month = year / 12

# Force (equivalents in Newtons)
newton = 1
lbf = slug * foot
pound_force = lbf

# Speed (equivalents in m/s)
kph = 1000 * meter / hour
knot = 1.852 * kph
mph = mile / hour
fps = foot / second

# Rotational Speed (equivalents in radians/second)
rads_per_sec = 1
rps = 2 * pi
rpm = rps / 60

# Volume (equivalents in m^3)
liter = 0.001
gallon_us = 231 * inch ** 3
gallon_imperial = 4.54609 * liter
gallon = gallon_us
quart = gallon_us / 4

# Pressure (equivalents in Pa)
pascal = 1
atm = 101325
torr = atm / 760
psi = lbf / inch ** 2
psf = lbf / foot ** 2

# Power (equivalents in Watts)
watt = 1
horsepower = 550 * foot * lbf / second
hp = horsepower

# Current (equivalents in Amperes)
amp = 1

# Energy (equivalents in Joules)
joule = 1
btu = 1055.05585262
calorie = 4.184
kcal = 1000 * calorie
watt_hour = watt * hour

# SI prefixes
yocto = 1e-24
zepto = 1e-21
atto = 1e-18
femto = 1e-15
pico = 1e-12
nano = 1e-9
micro = 1e-6
milli = 1e-3
centi = 1e-2
deci = 1e-1
deka = 1e1
hecto = 1e2
kilo = 1e3
mega = 1e6
giga = 1e9
tera = 1e12
peta = 1e15
exa = 1e18
zetta = 1e21
yotta = 1e24

# Some abbreviations / alternate names
m = meter
ft = foot
yd = yard
mi = mile
nmi = naut_mile
kilogram = kg
sec = second
s = second
min = minute
hr = hour
wk = week
yr = year
N = newton
kt = knot
L = liter
Pa = pascal



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_racecar(
        N=100,
        plot=False
):
    opti = asb.Opti()  # Optimization problem

    t_final = opti.variable(init_guess=1, lower_bound=0)
    t = np.linspace(0, t_final, N)

    x = opti.variable(init_guess=np.linspace(0, 1, N))
    v = opti.derivative_of(
        x, with_respect_to=t,
        derivative_init_guess=1,
        method="cubic"
    )
    u = opti.variable(init_guess=np.ones(N), lower_bound=0, upper_bound=1)
    opti.constrain_derivative(
        u - v,
        variable=v, with_respect_to=t,
        method="cubic"
    )

    from aerosandbox.numpy.integrate_discrete import (
        integrate_discrete_intervals,
        integrate_discrete_squared_curvature
    )

    effort = 0
    effort = 1e-6 * np.sum(
        integrate_discrete_squared_curvature(
            f=u,
            x=t
        )
    )

    opti.minimize(t_final + effort)

    opti.subject_to([
        v <= 1 - np.sin(2 * np.pi * x) / 2,
        x[0] == 0,
        v[0] == 0,
        x[-1] == 1,
    ])

    sol = opti.solve(
        behavior_on_failure="return_last"
    )
    print(f"t_final: {sol(t_final)}")
    print(f"error: {np.abs(1.9065661561917042 - sol(t_final))}")
    assert sol(t_final) == pytest.approx(1.9065661561917042, rel=1e-3)

    if plot:
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots()
        ax.plot(sol(t), sol(v), label="speed")
        ax.plot(sol(t), sol(x), label="pos")
        ax.plot(sol(t), 1 - np.sin(2 * np.pi * sol(x)) / 2, "r--", label="speed limit")
        ax.plot(sol(t), sol(u), "k", label="throttle")
        plt.ylim(0, 1.6)
        p.show_plot()


if __name__ == '__main__':

    N = 100  # number of control intervals

    opti = asb.Opti()  # Optimization problem

    t_final = opti.variable(init_guess=1, lower_bound=0)
    t = np.linspace(0, t_final, N)

    x = opti.variable(init_guess=np.linspace(0, 1, N))
    v = opti.derivative_of(
        x, with_respect_to=t,
        derivative_init_guess=1,
        method="cubic"
    )
    u = opti.variable(init_guess=np.ones(N), lower_bound=0, upper_bound=1)
    opti.constrain_derivative(
        u - v,
        variable=v, with_respect_to=t,
        method="cubic"
    )

    from aerosandbox.numpy.integrate_discrete import (
        integrate_discrete_intervals,
        integrate_discrete_squared_curvature
    )

    effort = 0
    effort = 1e-6 * np.sum(
        integrate_discrete_squared_curvature(
            f=u,
            x=t
        )
    )

    opti.minimize(t_final + effort)

    opti.subject_to([
        v <= 1 - np.sin(2 * np.pi * x) / 2,
        x[0] == 0,
        v[0] == 0,
        x[-1] == 1,
    ])

    sol = opti.solve(
        behavior_on_failure="return_last"
    )
    print(f"t_final: {sol(t_final)}")
    print(f"error: {np.abs(1.9065661561917042 - sol(t_final))}")

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()
    ax.plot(sol(t), sol(v), label="speed")
    ax.plot(sol(t), sol(x), label="pos")
    ax.plot(sol(t), 1 - np.sin(2 * np.pi * sol(x)) / 2, "r--", label="speed limit")
    ax.plot(sol(t), sol(u), ".-k", label="throttle")
    plt.ylim(0, 1.6)
    p.show_plot()



import aerosandbox.numpy as np
from aerosandbox.geometry.wing import Wing, WingXSec
from aerosandbox.geometry.airfoil.airfoil import Airfoil
from textwrap import indent, dedent
from aerosandbox.geometry.openvsp_io.asb_to_openvsp import _utilities


def generate_wing(
        wing: Wing,
        include_main=True
) -> str:
    """
    Generates a VSPScript file for a Wing object.

    Args:
        wing: The Wing object to generate a VSPScript file for.

    Returns: A VSPScript file as a string.

    """
    script = ""

    script += f"""\
//==== Add Wing "{wing.name}" ====//
string wid = AddGeom("WING");
ChangeXSecShape( GetXSecSurf( wid, 0 ), 0, XS_FILE_AIRFOIL );
ChangeXSecShape( GetXSecSurf( wid, 1 ), 1, XS_FILE_AIRFOIL );

"""

    script += """//==== Generate Blank Wing Sections ====//\n"""
    if len(wing.xsecs) < 2:
        raise ValueError("Wings must have at least 2 cross sections.")
    else:
        for i in range(len(wing.xsecs) - 1):
            script += f"""\
InsertXSec( wid, 1, XS_FILE_AIRFOIL ); // WingXSecs {i} to {i + 1}
"""

    script += f"""\
//===== Cut The Original Section =====//
CutXSec( wid, 1 );

//==== Set Overall Wing Options and First Section ====//
SetParmVal( wid, "X_Rel_Location", "XForm", {wing.xsecs[0].xyz_le[0]} );
SetParmVal( wid, "Y_Rel_Location", "XForm", {wing.xsecs[0].xyz_le[1]} );
SetParmVal( wid, "Z_Rel_Location", "XForm", {wing.xsecs[0].xyz_le[2]} );
SetParmVal( wid, "Twist", "XSec_0", {wing.xsecs[0].twist} );
SetParmVal( wid, "Twist_Location", "XSec_0", 0.0 );
SetParmVal( wid, "Sym_Ancestor", "Sym", 0.0 ); // Global Origin
SetParmVal( wid, "Sym_Planar_Flag", "Sym", {"2.0" if wing.symmetric else "0.0"} ); // 0 = None, 2 = XZ sym.
SetParmVal( wid, "RotateAirfoilMatchDideralFlag", "WingGeom", 1.0 ); // sic, typo from OpenVSP
Update();

"""

    script += """//==== Set Wing Section Options ====//\n"""
    xyz_le = np.stack([xsec.xyz_le for xsec in wing.xsecs], axis=0)
    dxyz_le = np.diff(xyz_le, axis=0)
    dx_le = dxyz_le[:, 0]
    dy_le = dxyz_le[:, 1]
    dz_le = dxyz_le[:, 2]
    dyz_le = np.sqrt(dy_le ** 2 + dz_le ** 2)
    dihedrals = np.arctan2d(dz_le, dy_le)
    sweep_le = np.arctan2d(dx_le, dyz_le)

    for i, (xsec_a, xsec_b) in enumerate(zip(wing.xsecs[:-1], wing.xsecs[1:])):
        script += f"""\
// ASB Section {i}, VSP Section {i + 1} (ASB XSecs {i} to {i + 1})
SetParmVal( wid, "Root_Chord", "XSec_{i + 1}", {xsec_a.chord} );
SetParmVal( wid, "Tip_Chord", "XSec_{i + 1}", {xsec_b.chord} );
SetParmVal( wid, "Span", "XSec_{i + 1}", {dyz_le[i]});
SetParmVal( wid, "Sweep", "XSec_{i + 1}", {sweep_le[i]} );
SetParmVal( wid, "Sweep_Location", "XSec_{i + 1}", 0.0 );
SetParmVal( wid, "Twist", "XSec_{i + 1}", {xsec_b.twist} );
SetParmVal( wid, "Twist_Location", "XSec_{i + 1}", 0.0 );
SetParmVal( wid, "Dihedral", "XSec_{i + 1}", {dihedrals[i]} );
SetParmVal( wid, "SectTess_U", "XSec_{i + 1}", 10 );
Update();

"""

    script += """\
//==== Set Airfoils ====//
"""

    for i in range(len(wing.xsecs)):
        xsec = wing.xsecs[i]

        upper = xsec.airfoil.upper_coordinates()[::-1]  # VSP wants front to back order
        lower = xsec.airfoil.lower_coordinates()  # Front to back order

        up_pnt_vecs = ", ".join([
            f"vec3d({p[0]:.8g}, {p[1]:.8g}, 0.0)" for p in upper
        ])
        lo_pnt_vecs = ", ".join([
            f"vec3d({p[0]:.8g}, {p[1]:.8g}, 0.0)" for p in lower
        ])

        script += f"""\
{{
    array<vec3d> up_pnt_vec = {{
        {up_pnt_vecs}
    }};
    array<vec3d> lo_pnt_vec = {{
        {lo_pnt_vecs}
    }};
    string xsec_surf = GetXSecSurf( wid, {i} );
    xsec_surf = GetXSecSurf( wid, {i} );
    string xsec = GetXSec( xsec_surf, {i} );
    SetParmVal( wid, "ThickChord", "XSecCurve_{i}", 0.5 ); // This is a hack that triggers VSP to update the curve
    SetParmVal( wid, "ThickChord", "XSecCurve_{i}", 1.0 );
    SetAirfoilPnts( xsec, up_pnt_vec, lo_pnt_vec );
    Update();
}}
"""

    script += """\
SetParmVal( wid, "Tess_W", "Shape", 100 );
Update();
"""

    if include_main:
        script = _utilities.wrap_script(script)

    return script


if __name__ == '__main__':
    wing = Wing(
        name="Main Wing",
        xsecs=[
            WingXSec(
                xyz_le=[1, 0, 0],
                chord=1.1,
                twist=-20,
                airfoil=Airfoil(name="dae11")
            ),
            WingXSec(
                xyz_le=[1, 2, 0],
                chord=0.9,
                twist=40,
                airfoil=Airfoil(name="NACA4412")
            ),
            WingXSec(
                xyz_le=[2, 5, 0],
                chord=0.5,
                twist=0,
                airfoil=Airfoil(name="NACA3412")
            ),
            WingXSec(
                xyz_le=[2, 10, 5],
                chord=0.25,
                twist=0,
                airfoil=Airfoil(name="NACA2412")
            )
        ]
    )
    print(generate_wing(wing))

    with open("test.vspscript", "w") as f:
        f.write(generate_wing(wing))



### This file contains an assortment of random airfoils to use
from aerosandbox.geometry.airfoil import Airfoil
from aerosandbox.library.aerodynamics.viscous import *
from aerosandbox.geometry.airfoil.airfoil_families import get_NACA_coordinates, \
    get_UIUC_coordinates


def diamond_airfoil(
        t_over_c: float,
        n_points_per_panel=2,
) -> Airfoil:
    x_nondim = [1, 0.5, 0, 0.5, 1]
    y_nondim = [0, 1, 0, -1, 0]

    x = np.concatenate(
        [
            list(np.cosspace(a, b, n_points_per_panel))[:-1]
            for a, b in zip(x_nondim[:-1], x_nondim[1:])
        ] + [[x_nondim[-1]]]
    )
    y = np.concatenate(
        [
            list(np.cosspace(a, b, n_points_per_panel))[:-1]
            for a, b in zip(y_nondim[:-1], y_nondim[1:])
        ] + [[y_nondim[-1]]]
    )
    y = y * (t_over_c / 2)

    coordinates = np.array([x, y]).T

    return Airfoil(
        name="Diamond",
        coordinates=coordinates,
    )




import aerosandbox.numpy as _np
import casadi as _cas
from typing import Tuple, Union


def softmax(
        *args: Union[float, _np.ndarray],
        softness: float = None,
        hardness: float = None,
) -> Union[float, _np.ndarray]:
    """
    An element-wise softmax between two or more arrays. Also referred to as the logsumexp() function.

    Useful for optimization because it's differentiable and preserves convexity!

    Great writeup by John D Cook here:
        https://www.johndcook.com/soft_maximum.pdf

    Notes: Can provide either `hardness` or `softness`, not both. These are the inverse of each other. If neither is
    provided, `hardness` is set to 1.

    Args:

        *args: Provide any number of arguments as values to take the softmax of.

        hardness: Hardness parameter. Higher values make this closer to max(x1, x2).

        softness: Softness parameter. (Inverse of hardness.) Lower values make this closer to max(x1, x2).

            - Setting `softness` is particularly useful, because it has the same units as each of the function's
            inputs. For example, if you're taking the softmax of two values that are lengths in units of meters,
            then `softness` is also in units of meters. In this case, `softness` has the rough meaning of "an amount
            of discrepancy between the input values that would be considered physically significant".

    Returns:
        Soft maximum of the supplied values.
    """
    ### Set defaults for hardness/softness
    n_specified_arguments = (hardness is not None) + (softness is not None)
    if n_specified_arguments == 0:
        softness = 1
    elif n_specified_arguments == 2:
        raise ValueError("You must provide exactly one of `hardness` or `softness`.")

    if hardness is not None:
        softness = 1 / hardness

    if _np.any(softness <= 0):
        if softness is not None:
            raise ValueError("The value of `softness` must be positive.")
        else:
            raise ValueError("The value of `hardness` must be positive.")

    if len(args) <= 1:
        raise ValueError("You must call softmax with the value of two or more arrays that you'd like to take the "
                         "element-wise softmax of.")

    ### Scale the args by softness
    args = [arg / softness for arg in args]

    ### Find the element-wise max and min of the arrays:
    min = args[0]
    max = args[0]
    for arg in args[1:]:
        min = _np.fmin(min, arg)
        max = _np.fmax(max, arg)

    out = max + _np.log(sum(
        [_np.exp(_np.maximum(array - max, -500)) for array in args]
    )
    )
    out = out * softness
    return out


def softmin(
        *args: Union[float, _np.ndarray],
        softness: float = None,
        hardness: float = None,
) -> Union[float, _np.ndarray]:
    """
    An element-wise softmin between two or more arrays. Related to the logsumexp() function.

    Useful for optimization because it's differentiable and preserves convexity!

    Great writeup by John D Cook here:
        https://www.johndcook.com/soft_maximum.pdf

    Notes: Can provide either `hardness` or `softness`, not both. These are the inverse of each other. If neither is
    provided, `hardness` is set to 1.

    Args:

        *args: Provide any number of arguments as values to take the softmin of.

        hardness: Hardness parameter. Higher values make this closer to min(x1, x2).

        softness: Softness parameter. (Inverse of hardness.) Lower values make this closer to min(x1, x2).

            - Setting `softness` is particularly useful, because it has the same units as each of the function's
            inputs. For example, if you're taking the softmin of two values that are lengths in units of meters,
            then `softness` is also in units of meters. In this case, `softness` has the rough meaning of "an amount
            of discrepancy between the input values that would be considered physically significant".

    Returns:
        Soft minimum of the supplied values.
    """
    return -softmax(
        *[-arg for arg in args],
        softness=softness,
        hardness=hardness,
    )


def softmax_scalefree(
        *args: Union[float, _np.ndarray],
        relative_softness: float = None,
        relative_hardness: float = None,
) -> Union[float, _np.ndarray]:
    n_specified_arguments = (relative_hardness is not None) + (relative_softness is not None)
    if n_specified_arguments == 0:
        relative_softness = 0.01
    elif n_specified_arguments == 2:
        raise ValueError("You must provide exactly one of `relative_softness` or `relative_hardness.")

    if relative_hardness is not None:
        relative_softness = 1 / relative_hardness

    return softmax(
        *args,
        softness=relative_softness * _np.linalg.norm(_np.array(args))
    )


def softmin_scalefree(
        *args: Union[float, _np.ndarray],
        relative_softness: float = None,
        relative_hardness: float = None,
) -> Union[float, _np.ndarray]:
    return -softmax_scalefree(
        *[-arg for arg in args],
        relative_softness=relative_softness,
        relative_hardness=relative_hardness,
    )


def softplus(
        x: Union[float, _np.ndarray],
        beta=1,
        threshold=40,
):
    """
    A smooth approximation of the ReLU function, applied elementwise to an array `x`.

    Softplus(x) = 1/beta * log(1 + exp(beta * x))

    Often used as an activation function in neural networks.

    Args:
        x: The input
        beta: A parameter that controls the "softness" of the function. Higher values of beta make the function
            approach ReLU.
        threshold: Values above this threshold are approximated as linear.

    Returns: The value of the softplus function.
    """
    if _np.is_casadi_type(x, recursive=False):
        return _np.where(
            beta * x > threshold,
            x,
            1 / beta * _cas.log1p(_cas.exp(beta * x))
        )
    else:
        return 1 / beta * _np.logaddexp(0, beta * x)


def sigmoid(
        x,
        sigmoid_type: str = "tanh",
        normalization_range: Tuple[Union[float, int], Union[float, int]] = (0, 1)
):
    """
    A sigmoid function. From Wikipedia (https://en.wikipedia.org/wiki/Sigmoid_function):
        A sigmoid function is a mathematical function having a characteristic "S"-shaped curve
        or sigmoid curve.

    Args:
        x: The input
        sigmoid_type: Type of sigmoid function to use [str]. Can be one of:
            * "tanh" or "logistic" (same thing)
            * "arctan"
            * "polynomial"
        normalization_type: Range in which to normalize the sigmoid, shorthanded here in the
            documentation as "N". This parameter is given as a two-element tuple (min, max).

            After normalization:
                >>> sigmoid(-Inf) == normalization_range[0]
                >>> sigmoid(Inf) == normalization_range[1]

            * In the special case of N = (0, 1):
                >>> sigmoid(-Inf) == 0
                >>> sigmoid(Inf) == 1
                >>> sigmoid(0) == 0.5
                >>> d(sigmoid)/dx at x=0 == 0.5
            * In the special case of N = (-1, 1):
                >>> sigmoid(-Inf) == -1
                >>> sigmoid(Inf) == 1
                >>> sigmoid(0) == 0
                >>> d(sigmoid)/dx at x=0 == 1

    Returns: The value of the sigmoid.
    """
    ### Sigmoid equations given here under the (-1, 1) normalization:
    if sigmoid_type == ("tanh" or "logistic"):
        # Note: tanh(x) is simply a scaled and shifted version of a logistic curve.
        s = _np.tanh(x)
    elif sigmoid_type == "arctan":
        s = 2 / _np.pi * _np.arctan(_np.pi / 2 * x)
    elif sigmoid_type == "polynomial":
        s = x / (1 + x ** 2) ** 0.5
    else:
        raise ValueError("Bad value of parameter 'type'!")

    ### Normalize
    min = normalization_range[0]
    max = normalization_range[1]
    s_normalized = s * (max - min) / 2 + (max + min) / 2

    return s_normalized


def swish(
        x,
        beta=1
):
    """
    A smooth approximation of the ReLU function, applied elementwise to an array `x`.

    Swish(x) = x / (1 + exp(-beta * x)) = x * logistic(x) = x * (0.5 + 0.5 * tanh(x/2))

    Often used as an activation function in neural networks.

    Args:
        x: The input
        beta: A parameter that controls the "softness" of the function. Higher values of beta make the function
            approach ReLU.

    Returns: The value of the swish function.
    """
    return x / (1 + _np.exp(-beta * x))


def blend(
        switch: float,
        value_switch_high,
        value_switch_low,
):
    """
    Smoothly blends between two values on the basis of some switch function.

    This function is similar in usage to numpy.where (documented here:
    https://numpy.org/doc/stable/reference/generated/numpy.where.html) , except that
    instead of using a boolean as to switch between the two values, a float is used to
    smoothly transition between the two in a differentiable manner.

    Before using this function, be sure to understand the difference between this and
    smoothmax(), and choose the correct one.

    Args:
        switch: A value that acts as a "switch" between the two values [float].
            If switch is -Inf, value_switch_low is returned.
            If switch is Inf, value_switch_high is returned.
            If switch is 0, the mean of value_switch_low and value_switch_high is returned.
            If switch is 1, the return value is roughly (0.88 * value_switch_high + 0.12 * value_switch_low).
            If switch is -1, the return value is roughly (0.88 * value_switch_low + 0.12 * value_switch_high).
        value_switch_high: Value to be returned when switch is high. Can be a float or an array.
        value_switch_low: Value to be returned when switch is low. Can be a float or an array.

    Returns: A value that is a blend between value_switch_low and value_switch_high, with the weighting dependent
        on the value of the 'switch' parameter.

    """
    blend_function = lambda x: sigmoid(
        x,
        normalization_range=(0, 1)
    )
    weight_to_value_switch_high = blend_function(switch)

    blend_value = (
            value_switch_high * weight_to_value_switch_high +
            value_switch_low * (1 - weight_to_value_switch_high)
    )

    return blend_value



import numpy as _onp
import casadi as _cas
from aerosandbox.numpy.determine_type import is_casadi_type
from aerosandbox.numpy.arithmetic_dyadic import centered_mod as _centered_mod
from aerosandbox.numpy.array import array, concatenate, reshape
from typing import Tuple


def diff(a, n=1, axis=-1, period=None):
    """
    Calculate the n-th discrete difference along the given axis.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.diff.html

    Adds one new argument, `period`, which is the period of the data. If provided, the difference is taken assuming
    the data "wraps around" at the period (i.e., modulo the period). For example:
    >>> diff([345, 355, 5, 15], period=360)
    >>> [10, 10, 10, 10]

    """
    if period is not None:
        return _centered_mod(
            diff(a, n=n, axis=axis),
            period
        )

    if not is_casadi_type(a):
        return _onp.diff(a, n=n, axis=axis)

    else:
        if axis != -1:
            raise NotImplementedError("This could be implemented, but haven't had the need yet.")

        result = a
        for i in range(n):
            result = _cas.diff(a)
        return result


def gradient(
        f,
        *varargs,
        axis=None,
        edge_order=1,
        n=1,
        period=None,
):
    """
    Return the gradient of an N-dimensional array.

    The gradient is computed using second order accurate central differences in the interior points and either first
    or second order accurate one-sides (forward or backwards) differences at the boundaries. The returned gradient
    hence has the same shape as the input array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.gradient.html

    Args:
        f: The array-like object to take the gradient of.
        *varargs: The spacing between the points of f. If a scalar, the spacing is assumed to be uniform in all
            dimensions. If an array, the array must have the same shape as f.

        axis: The axis along which the difference is taken. If None, the gradient is taken for all axes.

        edge_order: The order of the error at the boundaries. 1 means first order, 2 means second order.

        n: This is a new argument (not in NumPy) that specifies the order of the derivative to take. 1 is the first
            derivative (default), 2 is the second derivative. Doing `np.gradient(f, n=2)` results in less discretization
            error than doing `np.gradient(np.gradient(f))`.

        period: The period of the data. If provided, the gradient is taken assuming the data "wraps around" at the period
            (i.e., modulo the period). See `aerosandbox.numpy.diff()` for more information.


    Returns: The gradient of f.

    """
    if (
            not is_casadi_type(f)
            and all([not is_casadi_type(vararg) for vararg in varargs])
            and n == 1
            and period is None
    ):
        return _onp.gradient(
            f,
            *varargs,
            axis=axis,
            edge_order=edge_order
        )
    else:
        f = array(f)
        shape = f.shape

        # Handle the varargs argument
        if len(varargs) == 0:
            varargs = (1.,)

        if len(varargs) == 1:
            varargs = [varargs[0] for i in range(len(shape))]

        if len(varargs) != len(shape):
            raise ValueError("You must specify either 0, 1, or N varargs, where N is the number of dimensions of f.")
        else:
            dxes = []

            for i, vararg in enumerate(varargs):
                if _onp.prod(array(vararg).shape) == 1:  # If it's a scalar, you have dx values
                    dxes.append(
                        vararg * _onp.ones(shape[i] - 1)
                    )
                else:
                    dxes.append(
                        diff(
                            vararg,
                        )
                    )

        # Handle the axis argument, with the edge case the CasADi arrays are always 2D
        if axis is None:
            if is_casadi_type(f, recursive=False) and shape[1] == 1:
                axis = 0
            elif len(shape) <= 1:
                axis = 0
            else:
                axis = tuple(_onp.arange(len(shape)))

        try:
            tuple(axis)  # See if axis is iterable
            axis_is_iterable = True
        except TypeError:
            axis_is_iterable = False

        if axis_is_iterable:
            return [
                gradient(
                    f,
                    varargs[axis_i],
                    axis=axis_i,
                    edge_order=edge_order,
                    n=n,
                    period=period,
                )
                for axis_i in axis
            ]

        else:
            # Check validity of axis
            if axis < 0:
                axis = len(shape) + axis
            if is_casadi_type(f, recursive=False):
                if axis not in [0, 1]:
                    raise ValueError("axis must be 0 or 1 for CasADi arrays.")

            dx = dxes[axis]
            dx_shape = [1] * len(shape)
            dx_shape[axis] = shape[axis] - 1
            dx = reshape(dx, dx_shape)

            def get_slice(slice_obj: slice) -> Tuple[slice]:
                slices = [slice(None)] * len(shape)
                slices[axis] = slice_obj
                return tuple(slices)

            hm = dx[get_slice(slice(None, -1))]
            hp = dx[get_slice(slice(1, None))]

            dfp = (
                    f[get_slice(slice(2, None))]
                    - f[get_slice(slice(1, -1))]
            )
            dfm = (
                    f[get_slice(slice(1, -1))]
                    - f[get_slice(slice(None, -2))]
            )

            if period is not None:
                dfp = _centered_mod(dfp, period)
                dfm = _centered_mod(dfm, period)

            if n == 1:
                grad_f = (
                                 hm ** 2 * dfp + hp ** 2 * dfm
                         ) / (
                                 hm * hp * (hm + hp)
                         )

                if edge_order == 1:
                    # First point
                    df_f = dfm[get_slice(slice(0, 1))]
                    h_f = hm[get_slice(slice(0, 1))]
                    grad_f_first = df_f / h_f

                    # Last point
                    df_l = dfp[get_slice(slice(-1, None))]
                    h_l = hp[get_slice(slice(-1, None))]
                    grad_f_last = df_l / h_l

                elif edge_order == 2:
                    # First point
                    dfm_f = dfm[get_slice(slice(0, 1))]
                    dfp_f = dfp[get_slice(slice(0, 1))]
                    hm_f = hm[get_slice(slice(0, 1))]
                    hp_f = hp[get_slice(slice(0, 1))]
                    grad_f_first = (
                                           2 * dfm_f * hm_f * hp_f + dfm_f * hp_f ** 2 - dfp_f * hm_f ** 2
                                   ) / (
                                           hm_f * hp_f * (hm_f + hp_f)
                                   )

                    # Last point
                    dfm_l = dfm[get_slice(slice(-1, None))]
                    dfp_l = dfp[get_slice(slice(-1, None))]
                    hm_l = hm[get_slice(slice(-1, None))]
                    hp_l = hp[get_slice(slice(-1, None))]
                    grad_f_last = (
                                          -dfm_l * hp_l ** 2 + dfp_l * hm_l ** 2 + 2 * dfp_l * hm_l * hp_l
                                  ) / (
                                          hm_l * hp_l * (hm_l + hp_l)
                                  )

                else:
                    raise ValueError("Invalid edge_order.")

                grad_f = concatenate((
                    grad_f_first,
                    grad_f,
                    grad_f_last
                ), axis=axis)

                return grad_f

            elif n == 2:
                grad_grad_f = (
                        2 / (hm + hp) * (
                        dfp / hp - dfm / hm
                )
                )

                grad_grad_f_first = grad_grad_f[get_slice(slice(0, 1))]
                grad_grad_f_last = grad_grad_f[get_slice(slice(-1, None))]

                grad_grad_f = concatenate((
                    grad_grad_f_first,
                    grad_grad_f,
                    grad_grad_f_last
                ), axis=axis)

                return grad_grad_f

            else:
                raise ValueError(
                    "A second-order reconstructor only supports first derivatives (n=1) and second derivatives (n=2).")


def trapz(x, modify_endpoints=False):  # TODO unify with NumPy trapz, this is different
    """
    Computes each piece of the approximate integral of `x` via the trapezoidal method with unit spacing.
    Can be viewed as the opposite of diff().

    Args:
        x: The vector-like object (1D np.ndarray, cas.MX) to be integrated.

    Returns: A vector of length N-1 with each piece corresponding to the mean value of the function on the interval
        starting at index i.

    """
    import warnings
    warnings.warn(
        "trapz() will eventually be deprecated, since NumPy plans to remove it in the upcoming NumPy 2.0 release (2024). \n"
        "For discrete intervals, use asb.numpy.integrate_discrete_intervals(f, method=\"trapz\") instead.",
        PendingDeprecationWarning)

    integral = (
                       x[1:] + x[:-1]
               ) / 2
    if modify_endpoints:
        integral[0] = integral[0] + x[0] * 0.5
        integral[-1] = integral[-1] + x[-1] * 0.5

    return integral


if __name__ == '__main__':
    import aerosandbox as asb
    import aerosandbox.numpy as np

    import casadi as cas

    # print(diff(cas.DM([355, 5]), period=360))

    print(
        gradient(
            np.linspace(45, 55, 11) % 50,
            period=50
        )
    )

    #
    # # a = np.linspace(-500, 500, 21) % 360 - 180
    # # print(diff(a, period=360))
    #
    # x = np.cumsum(np.arange(10))
    # y = x ** 2
    #
    # print(gradient(y, x, edge_order=1))
    # print(gradient(y, x, edge_order=1))
    # print(gradient(y, x, edge_order=1, n=2))
    #
    # opti = asb.Opti()
    # x = opti.variable(init_guess=[355, 5])
    # d = diff(x, period=360)
    # opti.subject_to([
    #     # x[0] == 3,
    #     x[0] > 180,
    #     x[1] < 180,
    #     d < 20,
    #     d > -20
    # ])
    # opti.maximize(np.sum(np.cosd(x)))
    # sol = opti.solve(
    #     behavior_on_failure="return_last"
    # )
    # print(sol(x))



import aerosandbox.numpy as np


def overall_pressure_ratio_turboshaft_technology_limit(
        mass_turboshaft: float
) -> float:
    """
    Estimates the maximum-practically-achievable overall pressure ratio (OPR) of a turboshaft engine, as a function
    of its mass. A regression to historical data.

    Based on an envelope of data for both civilian and military turboshafts (including RC-scale turboshafts), available in:
    `aerosandbox/library/datasets/turbine_engines/data.xlsx`

    See study in `/AeroSandbox/studies/TurboshaftStudies/make_fit_overall_pressure_ratio.py` for model details.

    Args:

        mass_turboshaft: The mass of the turboshaft engine. [kg]

    Returns:

        The maximum-practically-achievable overall pressure ratio (OPR) of the turboshaft engine. [-]
    """
    p = {'scl': 1.0222956615376533, 'cen': 1.6535195257959798, 'high': 23.957335474997656}
    return np.blend(
        np.log10(mass_turboshaft) / p["scl"] - p["cen"],
        value_switch_high=p["high"],
        value_switch_low=1,
    )


def power_turboshaft(
        mass_turboshaft: float,
        overall_pressure_ratio: float = None,
) -> float:
    """
    Estimates the maximum rated power of a turboshaft engine, given its mass. A regression to historical data.

    Based on data for both civilian and military turboshafts, available in:
    `aerosandbox/library/datasets/turbine_engines/data.xlsx`

    See studies in `/AeroSandbox/studies/TurboshaftStudies/make_fit_power.py` for model details.

    Args:

        mass_turboshaft: The mass of the turboshaft engine. [kg]

        overall_pressure_ratio: The overall pressure ratio of the turboshaft engine. [-] If unspecified, a sensible
            default based on the technology envelope (with a 0.7x knockdown) will be used.

    Returns:

        The maximum (rated takeoff) power of the turboshaft engine. [W]

    """
    if overall_pressure_ratio is None:
        overall_pressure_ratio = overall_pressure_ratio_turboshaft_technology_limit(
            mass_turboshaft
        ) * 0.7

    p = {'a': 1674.9411795202134, 'OPR': 0.5090953411025091, 'Weight [kg]': 0.9418482117552568}
    return (
            p["a"]
            * mass_turboshaft ** p["Weight [kg]"]
            * overall_pressure_ratio ** p["OPR"]
    )


def thermal_efficiency_turboshaft(
        mass_turboshaft: float,
        overall_pressure_ratio: float = None,
        throttle_setting: float = 1,
) -> float:
    """
    Estimates the thermal efficiency of a turboshaft engine. A regression to historical data.

    Based on data for both civilian and military turboshafts, available in:
    `aerosandbox/library/datasets/turbine_engines/data.xlsx`

    See studies in `/AeroSandbox/studies/TurboshaftStudies/make_turboshaft_fits.py` for model details.

    Thermal efficiency knockdown at partial power is based on:
        Ingmar Geiß, "Sizing of the Series Hybrid-electric Propulsion System of General Aviation Aircraft", 2020.
        PhD Thesis, University of Stuttgart. Page 18, Figure 3.2.

    Args:

        mass_turboshaft: The mass of the turboshaft engine. [kg]

        overall_pressure_ratio: The overall pressure ratio of the turboshaft engine. [-] If unspecified, a sensible
            default based on the technology envelope (with a 0.7x knockdown) will be used.

        throttle_setting: The throttle setting of the turboshaft engine. [-] 1 is full throttle, 0 is no throttle.

    Returns:

        The thermal efficiency of the turboshaft engine. [-]

    """
    if overall_pressure_ratio is None:
        overall_pressure_ratio = overall_pressure_ratio_turboshaft_technology_limit(
            mass_turboshaft
        ) * 0.7

    p = {'a': 0.12721246565294902, 'wcen': 2.679474077211383, 'wscl': 4.10824884208311}

    ideal_efficiency = 1 - (1 / overall_pressure_ratio) ** ((1.4 - 1) / 1.4)

    thermal_efficiency_at_full_power = np.blend(
        p["a"] + (np.log10(mass_turboshaft) - p["wcen"]) / p["wscl"],
        value_switch_high=ideal_efficiency,
        value_switch_low=0,
    )

    p = {
        'B0': 0.0592,  # Modified from Geiß thesis such that B values sum to 1 by construction. Orig: 0.05658
        'B1': 2.567,
        'B2': -2.612,
        'B3': 0.9858
    }

    thermal_efficiency_knockdown_from_partial_power = (
            p["B0"]
            + p["B1"] * throttle_setting
            + p["B2"] * throttle_setting ** 2
            + p["B3"] * throttle_setting ** 3
    )

    return (
            thermal_efficiency_at_full_power
            * thermal_efficiency_knockdown_from_partial_power
    )


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots()
    x = np.linspace(0, 1)
    plt.plot(
        x,
        thermal_efficiency_turboshaft(1000, throttle_setting=x) / thermal_efficiency_turboshaft(1000),
    )
    ax.xaxis.set_major_formatter(p.mpl.ticker.PercentFormatter(1))
    plt.xlim(0, 1)
    plt.ylim(0, 1)
    p.set_ticks(0.1, 0.05, 0.1, 0.05)

    p.show_plot(
        "Turboshaft: Thermal Efficiency at Partial Power",
        "Throttle Setting [-]",
        "Thermal Efficiency Knockdown relative to Design Point [-] $\eta / \eta_\mathrm{max}$"
    )

    ##### Do Weight/OPR Efficiency Plot #####

    fig, ax = plt.subplots()
    mass = np.geomspace(1e0, 1e4, 300)
    opr = np.geomspace(1, 100, 500)

    Mass, Opr = np.meshgrid(mass, opr)

    Mask = overall_pressure_ratio_turboshaft_technology_limit(Mass) > Opr

    cont, contf, cbar = p.contour(
        Mass,
        Opr,
        thermal_efficiency_turboshaft(Mass, Opr),
        mask=Mask,
        linelabels_format=lambda x: f"{x:.0%}",
        x_log_scale=True,
        colorbar_label="Thermal Efficiency [%]",
        cmap="turbo_r",
    )

    cbar.ax.yaxis.set_major_formatter(p.mpl.ticker.PercentFormatter(1, decimals=0))

    p.set_ticks(None, None, 5, 1)

    p.show_plot(
        "Turboshaft Model: Thermal Efficiency vs. Weight and OPR",
        "Engine Weight [kg]",
        "Overall Pressure Ratio [-]",
        dpi=300
    )



import numpy as _onp
import casadi as _cas
from typing import List, Tuple, Dict, Union, Sequence
from aerosandbox.numpy.determine_type import is_casadi_type


def array(array_like, dtype=None):
    """
    Initializes a new array. Creates a NumPy array if possible; if not, creates a CasADi array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.array.html
    """
    if is_casadi_type(array_like, recursive=False):  # If you were literally given a CasADi array, just return it
        # Handles inputs like cas.DM([1, 2, 3])
        return array_like

    elif not is_casadi_type(array_like,
                            recursive=True) or dtype is not None:
        # If you were given a list of iterables that don't have CasADi types:
        # Handles inputs like [[1, 2, 3], [4, 5, 6]]
        return _onp.array(array_like, dtype=dtype)

    else:
        # Handles inputs like [[opti_var_1, opti_var_2], [opti_var_3, opti_var_4]]
        def make_row(contents: List):
            try:
                return _cas.horzcat(*contents)
            except (TypeError, Exception):
                return contents

        return _cas.vertcat(
            *[
                make_row(row)
                for row in array_like
            ]
        )


def concatenate(arrays: Sequence, axis: int = 0):
    """
    Join a sequence of arrays along an existing axis. Returns a NumPy array if possible; if not, returns a CasADi array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.concatenate.html
    """

    if not is_casadi_type(arrays, recursive=True):
        return _onp.concatenate(arrays, axis=axis)

    else:
        if axis == 0:
            return _cas.vertcat(*arrays)
        elif axis == 1:
            return _cas.horzcat(*arrays)
        else:
            raise ValueError("CasADi-backend arrays can only be 1D or 2D, so `axis` must be 0 or 1.")


def stack(arrays: Sequence, axis: int = 0):
    """
    Join a sequence of arrays along a new axis. Returns a NumPy array if possible; if not, returns a CasADi array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.stack.html
    """
    if not is_casadi_type(arrays, recursive=True):
        return _onp.stack(arrays, axis=axis)

    else:
        ### Validate stackability
        for array in arrays:
            if is_casadi_type(array, recursive=False):
                if not array.shape[1] == 1:
                    raise ValueError("Can only stack Nx1 CasADi arrays!")
            else:
                if not len(array.shape) == 1:
                    raise ValueError("Can only stack 1D NumPy ndarrays alongside CasADi arrays!")

        if axis == 0 or axis == -2:
            return _cas.transpose(_cas.horzcat(*arrays))
        elif axis == 1 or axis == -1:
            return _cas.horzcat(*arrays)
        else:
            raise ValueError("CasADi-backend arrays can only be 1D or 2D, so `axis` must be 0 or 1.")


def hstack(arrays):
    if not is_casadi_type(arrays, recursive=True):
        return _onp.hstack(arrays)
    else:
        raise ValueError(
            "Use `np.stack()` or `np.concatenate()` instead of `np.hstack()` when dealing with mixed-backend arrays.")


def vstack(arrays):
    if not is_casadi_type(arrays, recursive=True):
        return _onp.vstack(arrays)
    else:
        raise ValueError(
            "Use `np.stack()` or `np.concatenate()` instead of `np.vstack()` when dealing with mixed-backend arrays.")


def dstack(arrays):
    if not is_casadi_type(arrays, recursive=True):
        return _onp.dstack(arrays)
    else:
        raise ValueError(
            "Use `np.stack()` or `np.concatenate()` instead of `np.dstack()` when dealing with mixed-backend arrays.")


def length(array) -> int:
    """
    Returns the length of an 1D-array-like object. An extension of len() with slightly different functionality.
    Args:
        array:

    Returns:

    """
    if not is_casadi_type(array, recursive=False):
        try:
            return len(array)
        except TypeError:
            return 1

    else:
        if array.shape[0] != 1:
            return array.shape[0]
        else:
            return array.shape[1]


def diag(v, k=0):
    """
    Extract a diagonal or construct a diagonal array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.diag.html
    """
    if not is_casadi_type(v, recursive=False):
        return _onp.diag(v, k=k)

    else:

        if 1 in v.shape:  # If v is a 1D array, construct a diagonal matrix
            if v.shape[0] == 1:
                v = v.T

            if k == 0:
                return _cas.diag(v)

            else:
                n = v.shape[0]
                res = type(v).zeros(n + abs(k), n + abs(k))
                for i in range(n):
                    if k >= 0:
                        res[i, i + k] = v[i]
                    else:
                        res[i - k, i] = v[i]
                return res

        elif v.shape[0] == v.shape[1]:  # If v is a square matrix, extract the diagonal

            n = v.shape[0]

            if k >= 0:
                return array([
                    v[i, i + k]
                    for i in range(n - k)
                ])
            else:
                return array([
                    v[i - k, i]
                    for i in range(n + k)
                ])

        else:
            raise NotImplementedError("Haven't yet added logic for non-square matrices.")


def roll(a, shift, axis: int = None):
    """
    Roll array elements along a given axis.

    Elements that roll beyond the last position are re-introduced at the first.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.roll.html

    Parameters
    ----------
    a : array_like

        Input array.

    shift : int or tuple of ints

        The number of places by which elements are shifted. If a tuple, then axis must be a tuple of the same size,
        and each of the given axes is shifted by the corresponding number. If an int while axis is a tuple of ints,
        then the same value is used for all given axes.

    axis : int or tuple of ints, optional

        Axis or axes along which elements are shifted. By default, the array is flattened before shifting,
        after which the original shape is restored.

    Returns
    -------
    res : ndarray
        Output array, with the same shape as a.

    """
    if not is_casadi_type(a, recursive=False):
        return _onp.roll(a, shift, axis=axis)
    else:
        if axis is None:
            a_flat = reshape(a, -1)
            result = roll(a_flat, shift, axis=0)
            return reshape(result, a.shape)
        elif isinstance(axis, int):
            shift = shift % a.shape[axis]  # shift can be negative
            if shift != 0:
                slice1 = [slice(None)] * 2
                slice1[axis] = slice(-shift, None)
                slice2 = [slice(None)] * 2
                slice2[axis] = slice(-shift)
                result = concatenate([a[tuple(slice1)], a[tuple(slice2)]], axis=axis)
            else:
                result = a
            return result
        elif isinstance(axis, tuple):
            result = a
            if not isinstance(shift, tuple):
                shift = (shift,) * len(axis)
            for ax, sh in zip(axis, shift):
                result = roll(result, sh, ax)
            return result
        else:
            raise ValueError("'axis' must be None, an integer or a tuple of integers")


def max(a, axis=None):
    """
    Return the maximum of an array or maximum along an axis.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.max.html
    """

    if not is_casadi_type(a, recursive=False):
        return _onp.max(
            a,
            axis=axis,
        )

    else:

        if axis is None:
            return _cas.mmax(a)

        if axis == 0:
            if a.shape[1] == 1:
                return _cas.mmax(a)
            else:
                return array([
                    _cas.mmax(a[:, i])
                    for i in range(a.shape[1])
                ])

        elif axis == 1:
            if a.shape[0] == 1:
                return _cas.mmax(a)
            else:
                return array([
                    _cas.mmax(a[i, :])
                    for i in range(a.shape[0])
                ])

        else:
            raise ValueError(f'Invalid axis {axis} for CasADi array.')


def min(a, axis=None):
    """
    Return the minimum of an array or minimum along an axis.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.min.html
    """

    if not is_casadi_type(a, recursive=False):
        return _onp.min(
            a=a,
            axis=axis,
        )

    else:

        if axis is None:
            return _cas.mmin(a)

        if axis == 0:
            if a.shape[1] == 1:
                return _cas.mmin(a)
            else:
                return array([
                    _cas.mmin(a[:, i])
                    for i in range(a.shape[1])
                ])

        elif axis == 1:
            if a.shape[0] == 1:
                return _cas.mmin(a)
            else:
                return array([
                    _cas.mmin(a[i, :])
                    for i in range(a.shape[0])
                ])

        else:
            raise ValueError(f'Invalid axis {axis} for CasADi array.')


def reshape(a, newshape, order='C'):
    """
    Gives a new shape to an array without changing its data.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.reshape.html
    """

    if not is_casadi_type(a, recursive=False):
        return _onp.reshape(a, newshape, order=order)
    else:
        if isinstance(newshape, int):
            newshape = (newshape, 1)

        elif len(newshape) == 1:
            newshape = (newshape[0], 1)

        elif len(newshape) == 2:
            newshape = tuple(newshape)

        elif len(newshape) > 2:
            raise ValueError("CasADi data types are limited to no more than 2 dimensions.")

        if order == "C":
            return _cas.reshape(a.T, newshape[::-1]).T
        elif order == "F":
            return _cas.reshape(a, newshape)
        else:
            raise NotImplementedError("Only C and F orders are supported.")


def ravel(a, order='C'):
    """
    Returns a contiguous flattened array.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.ravel.html
    """
    if not is_casadi_type(a, recursive=False):
        return _onp.ravel(a, order=order)
    else:
        return reshape(a, -1, order=order)


def tile(A, reps):
    """
    Construct an array by repeating A the number of times given by reps.

    See syntax here: https://numpy.org/doc/stable/reference/generated/numpy.tile.html
    """
    if not is_casadi_type(A, recursive=False):
        return _onp.tile(A, reps)
    else:
        if len(reps) == 1:
            return _cas.repmat(A, reps[0], 1)
        elif len(reps) == 2:
            return _cas.repmat(A, reps[0], reps[1])
        else:
            raise ValueError("Cannot have >2D arrays when using CasADi numeric backend!")


def zeros_like(a, dtype=None, order='K', subok=True, shape=None):
    """Return an array of zeros with the same shape and type as a given array."""
    if not is_casadi_type(a, recursive=False):
        return _onp.zeros_like(a, dtype=dtype, order=order, subok=subok, shape=shape)
    else:
        return _onp.zeros(shape=length(a))


def ones_like(a, dtype=None, order='K', subok=True, shape=None):
    """Return an array of ones with the same shape and type as a given array."""
    if not is_casadi_type(a, recursive=False):
        return _onp.ones_like(a, dtype=dtype, order=order, subok=subok, shape=shape)
    else:
        return _onp.ones(shape=length(a))


def empty_like(prototype, dtype=None, order='K', subok=True, shape=None):
    """Return a new array with the same shape and type as a given array."""
    if not is_casadi_type(prototype, recursive=False):
        return _onp.empty_like(prototype, dtype=dtype, order=order, subok=subok, shape=shape)
    else:
        return zeros_like(prototype)


def full_like(a, fill_value, dtype=None, order='K', subok=True, shape=None):
    """Return a full array with the same shape and type as a given array."""
    if not is_casadi_type(a, recursive=False):
        return _onp.full_like(a, fill_value, dtype=dtype, order=order, subok=subok, shape=shape)
    else:
        return fill_value * ones_like(a)


def assert_equal_shape(
        arrays: Union[List[_onp.ndarray], Dict[str, _onp.ndarray]],
) -> None:
    """
    Assert that all of the given arrays are the same shape. If this is not true, raise a ValueError.

    Args: arrays: The arrays to be evaluated.

            Can be provided as a:

                * List, in which case a generic ValueError is thrown

                * Dictionary consisting of name:array pairs for key:value, in which case the names are given in the ValueError.

    Returns: None. Throws an error if leng

    """
    try:
        names = arrays.keys()
        arrays = list(arrays.values())
    except AttributeError:
        names = None

    def get_shape(array):
        try:
            return array.shape
        except AttributeError:  # If it's a float/int
            return ()

    shape = get_shape(arrays[0])

    for array in arrays[1:]:
        if not get_shape(array) == shape:
            if names is None:
                raise ValueError("The given arrays do not have the same shape!")
            else:
                namelist = ", ".join(names)
                raise ValueError(f"The given arrays {namelist} do not have the same shape!")



import sympy as s
from sympy import init_printing
init_printing()

# "Backward" -> gets the integral from x2 to x3, by analogy to backward Euler

# Define the symbols
x1, x2, x3 = s.symbols('x1 x2 x3', real=True)
f1, f2, f3 = s.symbols('f1 f2 f3', real=True)

h = x3 - x2
hm = x2 - x1

q = s.symbols('q')  # Normalized space for a Bernstein basis.
# Mapping from x-space to q-space has x=x2 -> q=0, x=x3 -> q=1.

q1 = s.symbols('q1', real=True)
q2 = 0
q3 = 1

# Define the Bernstein basis polynomials
b1 = (1 - q) ** 2
b2 = 2 * q * (1 - q)
b3 = q ** 2

c1, c2, c3 = s.symbols('c1 c2 c3', real=True)

# Can solve for c2 and c3 exactly
c1 = f2
c3 = f3

f = c1 * b1 + c2 * b2 + c3 * b3

f1_cubic = f.subs(q, q1)#.simplify()

factors = [q1]
# factors = [f1, f2, f3, f4]

# Solve for c2 and c3
sol = s.solve(
    [
        f1_cubic - f1,
    ],
    [
        c2,
    ],
)
c2 = sol[c2].factor(factors).simplify()


f = c1 * b1 + c2 * b2 + c3 * b3

integral = (c1 + c2 + c3) / 3 # God I love Bernstein polynomials
# integral = s.simplify(integral)

integral = integral.factor(factors).simplify()

parsimony = len(str(integral))
print(s.pretty(integral, num_columns=100))
print(f"Parsimony: {parsimony}")


import aerosandbox.numpy as np
from aerosandbox.geometry.propulsor import Propulsor
from textwrap import indent, dedent
from aerosandbox.geometry.openvsp_io.asb_to_openvsp import _utilities


def generate_propulsor(
        propulsor: Propulsor,
        include_main=True
) -> str:
    """
    Generates a VSPScript file for a Wing object.

    Args:
        wing: The Wing object to generate a VSPScript file for.

    Returns: A VSPScript file as a string.

    """
    script = ""

    script += f"""\
//==== Add Propulsor "{propulsor.name}" ====//
string pid = AddGeom("PROP");

"""

    ### Compute the orientation of the propulsor
    desired_normal = np.array(propulsor.xyz_normal) / np.linalg.norm(propulsor.xyz_normal)

    if np.allclose(desired_normal, np.array([1, 0, 0]), atol=1e-8):
        y_rot_deg = 180
        z_rot_deg = 0
    else:
        import aerosandbox as asb
        opti = asb.Opti()
        y_rot = opti.variable(init_guess=0, lower_bound=-np.pi, upper_bound=np.pi)
        z_rot = opti.variable(init_guess=0, lower_bound=-np.pi, upper_bound=np.pi)

        rot = (
                np.rotation_matrix_3D(angle=y_rot, axis="y") @
                np.rotation_matrix_3D(angle=z_rot, axis="z")
        )
        normal = rot @ np.array([-1, 0, 0])

        opti.maximize(
            np.dot(normal, desired_normal)
        )
        sol = opti.solve(verbose=False)
        y_rot_deg = np.degrees(sol(y_rot))
        z_rot_deg = np.degrees(sol(z_rot))

    script += f"""\
//==== Set Overall Propulsor Options and First Section ====//
SetParmVal( pid, "Y_Rel_Rotation", "XForm", {y_rot_deg:.8g} );
SetParmVal( pid, "Z_Rel_Rotation", "XForm", {z_rot_deg:.8g} );
SetParmVal( pid, "X_Rel_Location", "XForm", {propulsor.xyz_c[0]} );
SetParmVal( pid, "Y_Rel_Location", "XForm", {propulsor.xyz_c[1]} );
SetParmVal( pid, "Z_Rel_Location", "XForm", {propulsor.xyz_c[2]} );
SetParmVal( pid, "Diameter", "Design", {2 * propulsor.radius} );
SetParmVal( pid, "PropMode", "Design", 1.0 ); // 0 = Blades, 1 = Both, 2 = Disk
Update();

"""

    script += """\
SetParmVal( pid, "Tess_U", "Shape", 25 );
SetParmVal( pid, "Tess_W", "Shape", 100 );
Update();
    """

    if include_main:
        script = _utilities.wrap_script(script)

    return script


if __name__ == '__main__':

    prop = Propulsor(
        name="Prop",
        xyz_c=np.array([0, 0, 0]),
        xyz_normal=np.array([1, 0.1, 0]),
        radius=0.6,
    )
    print(generate_propulsor(prop))

    with open("test.vspscript", "w") as f:
        f.write(generate_propulsor(prop))



import inspect
from typing import Callable, Any, Union, Optional


def black_box(
        function: Callable[[Any], float],
        n_in: int = None,
        n_out: int = 1,
        fd_method: str ='central',
        fd_step: Optional[float] = None,
        fd_step_iter: Optional[bool] = None,
) -> Callable[[Any], float]:
    """
    Wraps a function as a black box, allowing it to be used in AeroSandbox / CasADi optimization problems.

    Obtains gradients via finite differences. Assumes that the function's Jacobian is fully dense, always.

    Args:

        function:

        n_in:

        n_out:

        fd_method: One of:
            - 'forward'
            - 'backward'
            - 'central'
            - 'smoothed'

    Returns:

    """
    ### Grab the signature of the function to be wrapped - we'll need it.
    signature = inspect.signature(function)

    ### Handle default arguments.
    if n_in is None:
        n_in = len(signature.parameters)

    if n_out is None:
        n_out = 1

    ### Add limitations
    if n_out > 1:
        raise NotImplementedError("Black boxes with multiple outputs are not yet supported.")

    ### Compute finite-differencing options
    fd_options = {}

    if fd_step is not None:
        fd_options['h'] = fd_step

    if fd_step_iter is not None:
        fd_options['h_iter'] = fd_step_iter

    import casadi as cas

    class BlackBox(cas.Callback):
        def __init__(
                self,
        ):
            cas.Callback.__init__(self)
            self.construct(
                self.__class__.__name__,
                dict(
                    enable_fd=True,
                    fd_method=fd_method,
                    fd_options=fd_options,
                )
            )

        # Number of inputs and outputs
        def get_n_in(self):
            """
            Number of scalar inputs to the black-box function.
            """
            return n_in

        def get_n_out(self):
            return n_out

        # Evaluate numerically
        def eval(self, args):
            f = function(*args)
            if isinstance(f, tuple):
                return f
            else:
                return [f]

    # `wrapped_function` is a function with the same call signature as the original function, but with all arguments as positional arguments.
    wrapped_function = BlackBox()

    def wrapped_function_with_kwargs_support(*args, **kwargs):
        """
        This is a function with the same call signature as the original function, allowing both positional and keyword
        arguments. Should work identically to the original function:
            - Keyword arguments should be optional, and the default values should be the same as the original function.
            - Keyword arguments should have the option of being passed as positional arguments, provided they are in the
                correct order and all required positional arguments are passed.
            - Keyword arguments should be allowed to be passed in any order.
            - Positional arguments should be required.
            - Positional arguments should have the option of being passed as keyword arguments.
        """
        inputs = []

        # Check number of positional arguments in the signature
        n_positional_args = len(signature.parameters) - len(signature.parameters.values())
        n_args = len(signature.parameters)
        if len(args) < n_positional_args or len(args) > n_args:
            raise TypeError(
                f"Takes from {n_positional_args} to {n_args} positional arguments but {len(args)} were given"
            )

        for i, (name, parameter) in enumerate(signature.parameters.items()):

            if i < len(args):
                input = args[i]
                if name in kwargs:
                    raise TypeError(
                        f"Got multiple values for argument '{name}': {input} and {kwargs[name]}"
                    )

            elif name in kwargs:
                input = kwargs[name]

            else:
                if parameter.default is parameter.empty:
                    raise TypeError(
                        f"Missing required argument '{name}'"
                    )
                else:
                    input = parameter.default

            # print(input)

            inputs.append(input)

        return wrapped_function(*inputs)

    wrapped_function_with_kwargs_support.wrapped_function = wrapped_function
    wrapped_function_with_kwargs_support.wrapper_class = BlackBox

    return wrapped_function_with_kwargs_support


if __name__ == '__main__':

    ### Create a function that's effectively black-box (doesn't use `aerosandbox.numpy`)
    def my_func(
            a1,
            a2,
            k1=4,
            k2=5,
            k3=6,
    ):
        import math
        return (
                math.sin(a1) * math.exp(a2) * math.cos(k1 * k2) + k3
        )


    ### Now, start an optimization problem
    import aerosandbox as asb
    import aerosandbox.numpy as np

    opti = asb.Opti()

    # Wrap our function such that it can be used in an optimization problem.
    my_func_wrapped = black_box(
        function=my_func, fd_method="central"
    )

    # Pick some variables to optimize over
    m = opti.variable(init_guess=5, lower_bound=3, upper_bound=8)
    n = opti.variable(init_guess=5, lower_bound=3, upper_bound=8)

    # Minimize the black-box function
    opti.minimize(
        my_func_wrapped(m, a2=3, k2=n)
    )

    # Solve
    sol = opti.solve()

    ### Plot the function over its inputs
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    M, N = np.meshgrid(
        np.linspace(3, 8, 300),
        np.linspace(3, 8, 300),
    )
    fig, ax = plt.subplots()
    p.contour(
        M, N, np.vectorize(my_func)(M, a2=3, k2=N),
    )
    p.show_plot()



import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.geometry.polygon import Polygon
import pytest


def test_polygon_creation():
    p = Polygon(
        coordinates=np.array([
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1]
        ])
    )
    assert p.n_points() == 4
    assert np.all(p.x() == np.array([0, 1, 1, 0]))
    assert np.all(p.y() == np.array([0, 0, 1, 1]))


def test_contains_points():
    p = Polygon(
        coordinates=np.array([
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1]
        ])
    )

    assert p.contains_points(0.5, 0.5) == True
    assert p.contains_points(-0.1, 0.5) == False
    assert p.contains_points(0.5, -0.1) == False
    assert p.contains_points(-0.1, -0.1) == False
    assert p.contains_points(1.1, 1.1) == False
    assert p.contains_points(1.0, 1.0) == True
    assert p.contains_points(0.5, 1.0) == True
    assert p.contains_points(0.5, 1.1) == False

    assert np.all(p.contains_points(
        x=np.array([0.5, 0.5, -0.1, -0.1]),
        y=np.array([0.5, -0.1, 0.5, -0.1])
    ) == np.array([True, False, False, False]))

    shape = (1, 2, 3, 4)
    x_points = np.random.randn(*shape)
    y_points = np.random.randn(*shape)
    contains = p.contains_points(x_points, y_points)
    assert shape == contains.shape


def test_equality():
    p1 = Polygon(
        coordinates=np.array([
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1]
        ])
    )

    p2 = p1.deepcopy()

    assert p1 == p2

    assert not p1 == p2.translate(0.1, 0)


def test_translate_scale_rotate():
    p1 = Polygon(
        coordinates=np.array([
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1]
        ])
    )
    p2 = (
        p1
        .translate(1, 1)
        .rotate(np.pi / 2, 1, 1)
        .scale(2, 1)
        .rotate(3 * np.pi / 2, 2, 1)
        .scale(1, 0.5)
        .translate(-2, -0.5)
    )

    assert np.allclose(
        p1.coordinates,
        p2.coordinates
    )


def test_jaccard_similarity():
    try:
        import shapely
    except ImportError:
        print("Shapely (optional) not installed; skipping test_jaccard_similarity.")
        return

    p1 = Polygon(
        coordinates=np.array([
            [0, 0],
            [1, 0],
            [1, 1],
            [0, 1]
        ])
    )
    p2 = p1.copy()

    assert p1.jaccard_similarity(p2) == pytest.approx(1)
    assert p1.jaccard_similarity(p2.translate(0.5, 0)) == pytest.approx(1 / 3)
    assert p1.jaccard_similarity(p2.translate(1, 0)) == pytest.approx(0)
    assert p1.jaccard_similarity(p2.translate(1.5, 0)) == pytest.approx(0)
    assert p1.jaccard_similarity(p2.translate(0.5, 0.5)) == pytest.approx(1 / 7)
    assert p1.jaccard_similarity(p2.translate(1, 1)) == pytest.approx(0)

    assert p1.jaccard_similarity(p2.rotate(np.pi / 2, 0.5, 0.5)) == pytest.approx(1)


if __name__ == '__main__':
    # test_translate_scale_rotate()
    pytest.main()



# from inviscid_analysis import N, gamma
import aerosandbox as asb
import aerosandbox.numpy as np

N = 1000
ue = np.linspace(1, 1, N)
x = np.linspace(0, 1, N)
nu = 1e-5
theta_0 = 1e-3
H_0 = 2.6

opti = asb.Opti()

theta = opti.variable(
    init_guess=theta_0,
    n_vars=N,
)
H = opti.variable(
    H_0,
    n_vars=N,
)

Re_theta = ue * theta / nu

H_star = np.where(
    H < 4,
    1.515 + 0.076 * (H - 4) ** 2 / H,
    1.515 + 0.040 * (H - 4) ** 2 / H
)  # From AVF Eq. 4.53
c_f = 2 / Re_theta * np.where(
    H < 6.2,
    -0.066 + 0.066 * np.abs(6.2 - H) ** 1.5 / (H - 1),
    -0.066 + 0.066 * (H - 6.2) ** 2 / (H - 4) ** 2
)  # From AVF Eq. 4.54
c_D = H_star / 2 / Re_theta * np.where(
    H < 4,
    0.207 + 0.00205 * np.abs(4 - H) ** 5.5,
    0.207 - 0.100 * (H - 4) ** 2 / H ** 2
)  # From AVF Eq. 4.55
Re_theta_o = 10 ** (
        2.492 / (H - 1) ** 0.43 +
        0.7 * (
                np.tanh(
                    14 / (H - 1) - 9.24
                ) + 1
        )
)  # From AVF Eq. 6.38

d_theta_dx = np.diff(theta) / np.diff(x)
d_ue_dx = np.diff(ue) / np.diff(x)
d_H_star_dx = np.diff(H_star) / np.diff(x)


def int(x):
    return (x[1:] + x[:-1]) / 2


def logint(x):
    # return int(x)
    logx = np.log(x)
    return np.exp(
        (logx[1:] + logx[:-1]) / 2
    )


### Add governing equations
opti.subject_to(
    d_theta_dx == int(c_f) / 2 - int(
        (H + 2) * theta / ue
    ) * d_ue_dx,  # From AVF Eq. 4.51
)
opti.subject_to(
    logint(theta / H_star) * d_H_star_dx ==
    int(2 * c_D / H_star - c_f / 2) +
    logint(
        (H - 1) * theta / ue
    ) * d_ue_dx
)

### Add initial conditions
opti.subject_to([
    theta[0] == theta_0,
    H[0] == H_0  # Equilibrium value
])

### Solve
sol = opti.solve()
theta = sol(theta)
H = sol(H)

### Plot
from aerosandbox.tools.pretty_plots import plt, show_plot

fig, ax = plt.subplots()
plt.plot(x, theta)
show_plot(r"$\theta$", "$x$", r"$\theta$")
fig, ax = plt.subplots()
plt.plot(x, H)
show_plot("$H$", "$x$", "$H$")



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_block_move_fixed_time():
    opti = asb.Opti()

    n_timesteps = 300

    time = np.linspace(0, 1, n_timesteps)

    dyn = asb.DynamicsPointMass1DHorizontal(
        mass_props=asb.MassProperties(mass=1),
        x_e=opti.variable(init_guess=np.linspace(0, 1, n_timesteps)),
        u_e=opti.variable(init_guess=1, n_vars=n_timesteps),
    )

    u = opti.variable(init_guess=np.linspace(1, -1, n_timesteps))

    dyn.add_force(
        Fx=u
    )

    dyn.constrain_derivatives(
        opti=opti,
        time=time
    )

    opti.subject_to([
        dyn.x_e[0] == 0,
        dyn.x_e[-1] == 1,
        dyn.u_e[0] == 0,
        dyn.u_e[-1] == 0,
    ])

    # effort = np.sum(
    #     np.trapz(dyn.X ** 2) * np.diff(time)
    # )

    effort = np.sum(  # More sophisticated integral-of-squares integration (closed form correct)
        np.diff(time) / 3 *
        (u[:-1] ** 2 + u[:-1] * u[1:] + u[1:] ** 2)
    )

    opti.minimize(effort)

    sol = opti.solve()

    dyn = sol(dyn)

    assert dyn.x_e[0] == pytest.approx(0)
    assert dyn.x_e[-1] == pytest.approx(1)
    assert dyn.u_e[0] == pytest.approx(0)
    assert dyn.u_e[-1] == pytest.approx(0)
    assert np.max(dyn.u_e) == pytest.approx(1.5, abs=0.01)
    assert sol(u)[0] == pytest.approx(6, abs=0.05)
    assert sol(u)[-1] == pytest.approx(-6, abs=0.05)


def test_block_move_minimum_time():
    opti = asb.Opti()

    n_timesteps = 300

    time = np.linspace(
        0,
        opti.variable(init_guess=1, lower_bound=0),
        n_timesteps,
    )

    dyn = asb.DynamicsPointMass1DHorizontal(
        mass_props=asb.MassProperties(mass=1),
        x_e=opti.variable(init_guess=np.linspace(0, 1, n_timesteps)),
        u_e=opti.variable(init_guess=1, n_vars=n_timesteps),
    )

    u = opti.variable(init_guess=np.linspace(1, -1, n_timesteps), lower_bound=-1, upper_bound=1)

    dyn.add_force(
        Fx=u
    )

    dyn.constrain_derivatives(
        opti=opti,
        time=time
    )

    opti.subject_to([
        dyn.x_e[0] == 0,
        dyn.x_e[-1] == 1,
        dyn.u_e[0] == 0,
        dyn.u_e[-1] == 0,
    ])

    opti.minimize(
        time[-1]
    )

    sol = opti.solve()

    dyn = sol(dyn)

    assert dyn.x_e[0] == pytest.approx(0)
    assert dyn.x_e[-1] == pytest.approx(1)
    assert dyn.u_e[0] == pytest.approx(0)
    assert dyn.u_e[-1] == pytest.approx(0)
    assert np.max(dyn.u_e) == pytest.approx(1, abs=0.01)
    assert sol(u)[0] == pytest.approx(1, abs=0.05)
    assert sol(u)[-1] == pytest.approx(-1, abs=0.05)
    assert np.mean(np.abs(sol(u))) == pytest.approx(1, abs=0.01)


if __name__ == '__main__':
    pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
from typing import List


def get_airfoil_database() -> List[asb.Airfoil]:
    airfoil_database_root = asb._asb_root / "geometry" / "airfoil" / "airfoil_database"

    afs = [
        asb.Airfoil(
            name=airfoil_file.stem,
        )
        for airfoil_file in airfoil_database_root.glob("*.dat")
    ]

    return afs

def test_repaneling_validity():
    try:
        import shapely
    except ModuleNotFoundError:
        pytest.skip("Shapely not installed; skipping this test.")

    afs = get_airfoil_database()

    for af in afs:
        try:
            similarity = af.jaccard_similarity(
                af.repanel(n_points_per_side=100)
            )
        except shapely.errors.GEOSException:
            continue # Jaccard similarity is not defined for self-intersecting polygons
            # similarity = np.nan
        assert similarity > 1 - 3 / af.n_points(), f"Airfoil {af.name} failed repaneling validity check with similarity {similarity}!"

def debug_draw(af: asb.Airfoil):
    if isinstance(af, str):
        af = asb.Airfoil(af)

    for af in [af, af.repanel()]:
        af.draw(
            draw_mcl=False, backend='plotly', show=False
        ).update_layout(
            yaxis=dict(scaleanchor=None)
        ).show()




if __name__ == '__main__':
    test_repaneling_validity()


import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_bounds():
    opti = asb.Opti()
    x = opti.variable(init_guess=3, log_transform=True, lower_bound=7)
    opti.minimize(x)
    sol = opti.solve()

    assert sol(x) == pytest.approx(7)


if __name__ == '__main__':
    pytest.main()
